
__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import os

from ...exceptions import *
from ..            import base
from ..cpi         import SYNC_CALL, ASYNC_CALL
from ..cpi         import context as cpi
from ...           import context as api
from ...constants  import *
from ...exceptions import *


######################################################################
#
# adaptor meta data
#
_ADAPTOR_NAME          = 'radical.saga.adaptors.ssh'
_ADAPTOR_SCHEMAS       = ['ssh']
_ADAPTOR_OPTIONS       = []

# FIXME: complete attribute list
_ADAPTOR_CAPABILITIES  = {
     'ctx_attributes'   : {TYPE      : "This MUST be set to ssh",
                           USER_ID   : "user name on target machine",
                           USER_KEY  : "maps to public ssh key",
                           USER_CERT : "maps to private ssh key",
                           USER_PASS : "passphrase for encrypted keys"}
}

_ADAPTOR_DOC           = {
    'name'             : _ADAPTOR_NAME,
    'cfg_options'      : _ADAPTOR_OPTIONS,
    'capabilities'     : _ADAPTOR_CAPABILITIES,
    'description'      : """

This SSH :class:`radical.saga.Context` adaptor points to an ssh keypair and
a user_id to be used for ssh based backend connections.  For example, an ssh
context can be used to start jobs (:class:`radical.saga.job.Job`) via ssh, to
copy files (:class:`radical.saga.filesystem.File`) via sftp, etc.

Not all supported attributes have to be defined when using an ssh context
adaptor -- unspecified attributes will have sensible default values.  For
example, the ``c.user_id`` will default to the local user id, and the default
passphrase in ``c.user_pass`` will be empty.

The `UserKey` and `UserCert` attributes can point to either the public or
private key of the ssh keypair -- the RADICAL-SAGA implementation will internally
complete the respective other key (public key file names are expected to be
derived from the private key, by appending the suffix `.pub` -- `.pem` files are
expected to contain both public and private key.).
    """,

    'schemas'          : {'ssh' : 'ssh key and userid information.'},
    'example'          : "examples/context/context_ssh.py"
}

_ADAPTOR_INFO          = {
    'name'             : _ADAPTOR_NAME,
    'version'          : 'v0.1',
    'schemas'          : _ADAPTOR_SCHEMAS,
    'cpis'             : [{
        'type'         : 'radical.saga.Context',
        'class'        : 'ContextSSH'
        }
    ]
}


# ------------------------------------------------------------------------------
#
class Adaptor (base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :

        base.Base.__init__ (self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        # there are no default myproxy contexts
        self._default_contexts = []
        self._have_defaults    = False


    # --------------------------------------------------------------------------
    #
    def sanity_check (self) :
        pass


    # --------------------------------------------------------------------------
    #
    def _get_default_contexts (self) :

        if not self._have_defaults :

            import glob
            candidate_certs = glob.glob ("%s/.ssh/*" % os.environ['HOME'])

            for key in candidate_certs :

                if  os.path.isdir (key) :
                    # don't want directories -- only keys
                    continue

                elif  key.endswith ('.pub') :
                    # don't want public keys in this loop
                    continue

                elif  key.endswith ('.pem') :
                    # we can't handle passwd protected PEMs well, since we can't
                    # detect when they are passwd protected in the first place,
                    # so we do not pick them up by default
                    continue
                else :
                    pub  = "%s.pub" % key

                # the private and public keys must exist
                if  not os.path.exists (key ) or \
                    not os.path.isfile (key )    :
                    self._logger.info ("ignore ssh key at %s (no private key: %s)" %  (key, key))
                    continue

                if  not os.path.exists (pub) or \
                    not os.path.isfile (pub)    :
                    self._logger.info ("ignore ssh key at %s (no public key: %s)" %  (key, pub))
                    continue


                try :
                    fh_key  = open (key )
                except Exception as e:
                    self._logger.info ("ignore ssh key at %s (key not readable: %s)" %  (key, e))
                    continue
                else :
                    fh_key .close ()


                try :
                    fh_pub  = open (pub )
                except Exception as e:
                    self._logger.info ("ignore ssh key at %s (public key %s not readable: %s)" %  (key, pub, e))
                    continue
                else :
                    fh_pub .close ()


                import subprocess
                if  not subprocess.call (["sh", "-c", "grep ENCRYPTED %s > /dev/null" % key]) :
                    # needs passphrase.  Great, actually, but won't work for
                    # default contexts as long as we can't prompt for pass
                    # phrases...
                    self._logger.warning ("ignore  ssh key at %s (requires passphrase)" %  key)
                    continue

                c = api.Context ('ssh')
                c.user_key  = key
                c.user_cert = pub

                self._default_contexts.append (c)

                self._logger.info ("default ssh key at %s" %  key)

            self._have_defaults = True


        # have defaults, and can return them...
        return self._default_contexts



# ------------------------------------------------------------------------------
#
class ContextSSH (cpi.Context) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        _cpi_base = super  (ContextSSH, self)
        _cpi_base.__init__ (api, adaptor)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, type) :

        if not type.lower () in (schema.lower() for schema in _ADAPTOR_SCHEMAS) :
            raise BadParameter \
                    ("the ssh context adaptor only handles ssh contexts - duh!")

        self.get_api ().type = type

        return self.get_api ()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def _initialize (self, session) :

        # make sure we have can access the key
        api = self.get_api ()

        unexpanded_key = None
        unexpanded_pub = None
        pwd = None

        if api.attribute_exists(USER_KEY ) :
            unexpanded_key = api.get_attribute(USER_KEY )
        if api.attribute_exists(USER_CERT) :
            unexpanded_pub = api.get_attribute(USER_CERT)
        if api.attribute_exists(USER_PASS) :
            pwd  = api.get_attribute(USER_PASS)

        # Expand any environment variables in the key/pub paths
        if unexpanded_key:
            key = os.path.expandvars(unexpanded_key)
        else:
            key = None
        if unexpanded_pub:
            pub = os.path.expandvars(unexpanded_pub)
        else:
            pub = None

        # either user_key or user_cert should be specified (or both),
        # then we complement the other, and convert to/from private
        # from/to public keys
        if  pub  and not key :
            key  = pub

        if  not key :
            # nothing to do, really.  This likely means that ssh setup is
            # done out-of-band.
            return

        # convert public key into private key
        if  key.endswith ('.pub') :
            if  not pub :
                pub = key
            key = key[:-4]
        elif key.endswith ('.pem') :
            if  not pub :
                pub = key
        else :
            if  not pub :
                pub = key+'.pub'

        # update the context with these setting
        api.set_attribute (USER_KEY , key)
        api.set_attribute (USER_CERT, pub)


        # the private and public keys must exist
        if  not os.path.exists (key) or \
            not os.path.isfile (key)    :
            raise BadParameter ("ssh key inaccessible: %s" % (key))

        if  not os.path.exists (pub) or \
            not os.path.isfile (pub)    :
            raise BadParameter ("ssh public key inaccessible: %s" % (pub))


        try :
            fh_key = open (key)
        except Exception as e:
            raise PermissionDenied ("ssh key '%s' not readable: %s" % (key, e)) \
                  from e
        else :
            fh_key.close ()


        try :
            fh_pub = open (pub)
        except Exception as e:
            raise PermissionDenied ("ssh public key '%s' not readable: %s" % (pub, e)) \
                  from e
        else :
            fh_pub.close ()


        import subprocess
        if  not subprocess.call (["sh", "-c", "grep ENCRYPTED %s > /dev/null" % key]) :
            if  pwd  :
                if  subprocess.call (["sh", "-c", "ssh-keygen -y -f %s -P '%s' > /dev/null" % (key, pwd)]) :
                    raise PermissionDenied ("ssh key '%s' is encrypted, incorrect password" % (key))
            else :
                self._logger.error ("ssh key '%s' is encrypted, unknown password" % (key))


        self._logger.info ("init SSH context for key  at '%s' done" % key)








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from ..namespace import constants as ns

# filesystem flags enum:
OVERWRITE      = ns.OVERWRITE       #      1
RECURSIVE      = ns.RECURSIVE       #      2
DEREFERENCE    = ns.DEREFERENCE     #      4
CREATE         = ns.CREATE          #      8
EXCLUSIVE      = ns.EXCLUSIVE       #     16
LOCK           = ns.LOCK            #     32
CREATE_PARENTS = ns.CREATE_PARENTS  #     64
TRUNCATE       =                         128
APPEND         =                         256
READ           =                         512
WRITE          =                        1024
READ_WRITE     =                        1536
BINARY         =                        2048

# filesystem seek_mode enum:
START          = "Start"
CURRENT        = "Current"
END            = "End"








__author__    = "Andre Merzky, Ole Weidner, Alexander Grill"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils            as ru
import radical.utils.signatures as rus


from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base    as sab

from .. import session          as ss
from .. import task             as st
from .  import constants        as c
from .  import entry


# ------------------------------------------------------------------------------
#
class Directory (entry.Entry) :
    '''
    Represents a SAGA directory as defined in GFD.90

    The saga.namespace.Directory class represents, as the name indicates,
    a directory on some (local or remote) namespace.  That class offers
    a number of operations on that directory, such as listing its contents,
    copying entries, or creating subdirectories::

        # get a directory handle
        dir = saga.namespace.Directory("sftp://localhost/tmp/")

        # create a subdir
        dir.make_dir ("data/")

        # list contents of the directory
        entries = dir.list ()

        # copy *.dat entries into the subdir
        for f in entries :
            if f ^ '^.*\.dat$' :
                dir.copy (f, "sftp://localhost/tmp/data/")


    Implementation note:
    ^^^^^^^^^^^^^^^^^^^^

    The SAGA API Specification (GFD.90) prescribes method overloading on method
    signatures, but that is not supported by Python (Python only does method
    overwriting).  So we implement one generic method version here, and do the
    case switching based on the provided parameter set.
    '''

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, flags=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        '''
        :param url: Url of the (remote) entry system directory.
        :type  url: :class:`saga.Url`

        flags:     flags enum
        session:   saga.Session
        ret:       obj

        Construct a new directory object

        The specified directory is expected to exist -- otherwise
        a DoesNotExist exception is raised.  Also, the URL must point to
        a directory (not to an entry), otherwise a BadParameter exception is
        raised.

        Example::

            # open some directory
            dir = saga.namespace.Directory("sftp://localhost/tmp/")

            # and list its contents
            entries = dir.list ()

        '''

        if  not flags : flags = 0
        self._nsentry = super  (Directory, self)
        self._nsentry.__init__ (url, flags, session,
                                _adaptor, _adaptor_state, _ttype=_ttype)


    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create (cls, url=None, flags=None, session=None, ttype=None) :
        '''
        url:       saga.Url
        flags:     saga.namespace.flags enum
        session:   saga.Session
        ttype:     saga.task.type enum
        ret:       saga.Task
        '''

        if  not flags : flags = 0
        _nsentry = super (Directory, cls)
        return _nsentry.create (url, flags, session, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((entry.Entry, st.Task))
    def open (self, name, flags=None, ttype=None) :
        '''
        name:     saga.Url
        flags:    saga.namespace.flags enum
        ttype:    saga.task.type enum
        ret:      saga.namespace.Entry / saga.Task
        '''
        if  not flags : flags = 0
        url = ru.Url(name)
        return self._adaptor.open (url, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (('Directory', st.Task))
    def open_dir (self, path, flags=None, ttype=None) :
        '''
        :param path: name/path of the directory to open
        :param flags: directory creation flags

        ttype:    saga.task.type enum
        ret:      saga.namespace.Directory / saga.Task

        Open and return a new directoy

           The call opens and returns a directory at the given location.

           Example::

               # create a subdir 'data' in /tmp
               dir = saga.namespace.Directory("sftp://localhost/tmp/")
               data = dir.open_dir ('data/', saga.namespace.Create)
        '''
        if  not flags : flags = 0
        return self._adaptor.open_dir (ru.Url(path), flags, ttype=ttype)


    # ----------------------------------------------------------------
    #
    # namespace directory methods
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def make_dir (self, tgt, flags=0, ttype=None) :
        '''
        :param tgt:   name/path of the new directory
        :param flags: directory creation flags

        ttype:         saga.task.type enum
        ret:           None / saga.Task

        Create a new directoy

        The call creates a directory at the given location.

        Example::

            # create a subdir 'data' in /tmp
            dir = saga.namespace.Directory("sftp://localhost/tmp/")
            dir.make_dir ('data/')
        '''
        if  not flags : flags = 0
        return self._adaptor.make_dir (ru.Url (tgt), flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def change_dir (self, url, flags=0, ttype=None) :
        '''
        url:           saga.Url
        flags:         flags enum
        ttype:         saga.task.type enum
        ret:           None / saga.Task
        '''
        if  not flags : flags = 0
        return self._adaptor.change_dir (url, flags=flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional (str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (ru.Url), st.Task))
    def list (self, pattern=None, flags=0, ttype=None) :
        '''
        :param pattern: Entry name pattern (like POSIX 'ls', e.g. '\*.txt')

        flags:         flags enum
        ttype:         saga.task.type enum
        ret:           list [saga.Url] / saga.Task

        List the directory's content

        The call will return a list of entries and subdirectories within the
        directory::

            # list contents of the directory
            for f in dir.list() :
                print(f)
        '''
        if  not flags : flags = 0
        return self._adaptor.list (pattern, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def exists (self, path, ttype=None) :

        '''
        :param path: path of the entry to check

        ttype:         saga.task.type enum
        ret:           bool / saga.Task

        Returns True if path exists, False otherwise.


        Example::

            # inspect an entry
            dir  = saga.namespace.Directory("sftp://localhost/tmp/")
            if dir.exists ('data'):
                # do something
        '''
        return self._adaptor.exists (path, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional (str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (ru.Url), st.Task))
    def find (self, pattern, flags=c.RECURSIVE, ttype=None) :
        '''
        pattern:       string
        flags:         flags enum
        ttype:         saga.task.type enum
        ret:           list [saga.Url] / saga.Task
        '''
        if  not flags : flags = 0
        return self._adaptor.find (pattern, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def get_num_entries (self, ttype=None) :
        '''
        ttype:         saga.task.type enum
        ret:           int / saga.Task
        '''
        return self._adaptor.get_num_entries (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  int,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((ru.Url, st.Task))
    def get_entry (self, num, ttype=None) :
        '''
        num:           int
        ttype:         saga.task.type enum
        ret:           saga.Url / saga.Task
        '''
        return self._adaptor.get_entry (num, ttype=ttype)


    # ----------------------------------------------------------------
    #
    # methods overloaded from namespace.Entry
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def copy (self, url_1, url_2=None, flags=0, ttype=None) :
        '''
        :param src: path of the entry to copy
        :param tgt: absolute URL of target name or directory

        url_1:         saga.Url
        url_2:         saga.Url / None
        flags:         flags enum / None
        ttype:         saga.task.type enum / None
        ret:           None / saga.Task

        Copy an entry from source to target

        The source is copied to the given target directory.  The path of the
        source can be relative::

            # copy an entry
            dir = saga.namespace.Directory("sftp://localhost/tmp/")
            dir.copy ("./data.bin", "sftp://localhost/tmp/data/")
        '''

        # FIXME: re-implement the url switching (commented out below)

        if not flags: flags = 0

        if url_2: return self._adaptor.copy(url_1, url_2, flags, ttype=ttype)
        else    : return self._nsentry.copy(url_1,        flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def link (self, url_1, url_2=None, flags=0, ttype=None) :
        '''
        src:           saga.Url
        tgt:           saga.Url
        flags:         flags enum
        ttype:         saga.task.type enum
        ret:           None / saga.Task
        '''
        if not flags: flags = 0

        if url_2: return self._adaptor.link(url_1, url_2, flags, ttype=ttype)
        else    : return self._nsentry.link(url_1,        flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  (ru.Url, str),
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def move (self, url_1, url_2=None, flags=0, ttype=None) :
        '''
        :param src: path of the entry to copy
        :param tgt: absolute URL of target directory

        flags:         flags enum
        ttype:         saga.task.type enum
        ret:           None / saga.Task

        Move an entry from source to target

        The source is moved to the given target directory.  The path of the
        source can be relative::

            # copy an entry
            dir = saga.namespace.Directory("sftp://localhost/tmp/")
            dir.move ("./data.bin", "sftp://localhost/tmp/data/")

        '''
        if not flags: flags = 0

        if url_2: return self._adaptor.move(url_1, url_2, flags, ttype=ttype)
        else    : return self._nsentry.move(url_1,        flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes('Directory',
               (ru.Url, str),
               rus.optional (int, rus.nothing),
               rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns((rus.nothing, st.Task))
    def remove (self, tgt=None, flags=0, ttype=None) :
        '''
        tgt:           saga.Url
        flags:         flags enum
        ttype:         saga.task.type enum
        ret:           None / saga.Task
        '''
        if not flags: flags = 0

        if tgt: return self._adaptor.remove(tgt, flags, ttype=ttype)
        else  : return self._nsentry.remove(     flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_dir (self, tgt=None, ttype=None) :
        '''
        tgt:           saga.Url / None
        ttype:         saga.task.type enum
        ret:           bool / saga.Task

        Returns True if path is a directory, False otherwise.

        Example::

            # inspect an entry
            dir  = saga.namespace.Directory("sftp://localhost/tmp/")
            if dir.is_dir ('data'):
                # do something
        '''
        if tgt: return self._adaptor.is_dir(tgt, ttype=ttype)
        else  : return self._nsentry.is_dir(     ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes('Directory',
               rus.optional ((ru.Url, str)),
               rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns((bool, st.Task))
    def is_entry(self, tgt=None, ttype=None) :
        '''
        tgt:           saga.Url / None
        ttype:         saga.task.type enum
        ret:           bool / saga.Task
        '''
        if tgt: return self._adaptor.is_entry(tgt, ttype=ttype)
        else  : return self._nsentry.is_entry(     ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_link (self, tgt=None, ttype=None) :
        '''
        tgt:           saga.Url / None
        ttype:         saga.task.type enum
        ret:           bool / saga.Task
        '''
        if tgt: return self._adaptor.is_link(tgt, ttype=ttype)
        else  : return self._nsentry.is_link(     ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Directory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((ru.Url, st.Task))
    def read_link (self, tgt=None, ttype=None) :
        '''
        tgt:           saga.Url / None
        ttype:         saga.task.type enum
        ret:           saga.Url / saga.Task
        '''

        if tgt: return self._adaptor.read_link(tgt, ttype=ttype)
        else  : return self._nsentry.read_link(     ttype=ttype)


# ------------------------------------------------------------------------------




# -*- coding: utf-8 -*-

__author__    = "Andre Merzky, Christian P.-Llamas, Ole Weidner, " \
                "Thomas Schatz, Alexander Grill"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" SGE job adaptor implementation
"""

import re
import os
import time

from datetime import datetime
from io import StringIO
from urllib.parse import parse_qs

import radical.utils as ru

from ...              import exceptions as rse
from ...              import utils      as rsu
from ...utils         import pty_shell  as rsups
from ...utils         import misc       as rsumisc
from ...              import job        as api
from ...adaptors      import base       as a_base
from ...adaptors.cpi  import job        as cpi
from ...adaptors.cpi  import decorators as cpi_decs
from ...job           import constants  as c


SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


_QSTAT_JOB_STATE_RE = re.compile(
    r"^([^ ]+) ([0-9]{2}/[0-9]{2}/[0-9]{4} [0-9]{2}:[0-9]{2}:[0-9]{2}) (.+)$")


class SgeKeyValueParser(object):
    """
    Parser for SGE commands returning lines with key-value pairs.
    It takes into account multi-line key-value pairs.
    It works as an iterator returning (key, value) tuples or as a dictionary.
    It allows to filter for keys.
    """

    KEY_VALUE_RE = re.compile(r"^([^ ]+) +(.+)$")

    def __init__(self, stream, filter_keys=None, key_suffix=None):
        """
        :param stream: an string or a file-like object implementing readline()
        :param filter_keys: an iterable with the list of keys of interest.
        :param key_suffix: a key suffix to remove when parsing
        """

        # check whether it is an string or a file-like object
        if isinstance(stream, str):
            self.stream = StringIO(stream)
        else:
            self.stream = stream

        self.filter_keys = set(filter_keys) if filter_keys is not None else None
        self.key_suffix = key_suffix

    def __next__(self):
        """
        Return the next key-value pair.
        :return: (key, value)
        """

        key, value = None, None
        while key is None:
            line = self.stream.readline()
            if len(line) == 0:
                raise StopIteration

            line = line.rstrip(" \n")

            # check for multi-line options
            while len(line) > 0 and line[-1] == "\\":
                line = line[:-1] + self.stream.readline() \
                                              .rstrip(" \n") \
                                              .lstrip(" ")

            m = self.KEY_VALUE_RE.match(line)
            if m is not None:

                key, value = m.groups()

                if  self.key_suffix is not None and \
                    key.endswith(self.key_suffix):
                    key = key[:-len(self.key_suffix)]

                if  self.filter_keys is not None and \
                    key not in self.filter_keys:
                    key = None  # skip this pair

        return key, value


    def __iter__(self):
        return self


    def as_dict(self):
        """
        Parses the key-value pairs and return them as a dictionary.
        :return: a dictionary containing key-value pairs parsed from
                 a SGE command.
        """

        d = dict()
        for key, value in self:
            d[key] = value

        return d


# --------------------------------------------------------------------
# some private defs
#
_PTY_TIMEOUT = 2.0

# --------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.sgejob"
_ADAPTOR_SCHEMAS       = ["sge", "sge+ssh", "sge+gsissh"]

# --------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES = {
    "jdes_attributes":   [c.NAME,
                          c.EXECUTABLE,
                          c.ARGUMENTS,
                          c.ENVIRONMENT,
                          c.INPUT,
                          c.OUTPUT,
                          c.ERROR,
                          c.QUEUE,
                          c.PROJECT,
                          c.WALL_TIME_LIMIT,
                          c.WORKING_DIRECTORY,
                          c.SPMD_VARIATION,
                          c.TOTAL_CPU_COUNT,
                          c.PROCESSES_PER_HOST,
                          c.CANDIDATE_HOSTS,
                          c.TOTAL_PHYSICAL_MEMORY],
    "job_attributes":    [c.EXIT_CODE,
                          c.EXECUTION_HOSTS,
                          c.CREATED,
                          c.STARTED,
                          c.FINISHED],
    "metrics":           [c.STATE],
    "contexts":          {"ssh": "SSH public/private keypair",
                          "x509": "GSISSH X509 proxy context",
                          "userpass": "username/password pair (ssh)"}
}

# --------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC = {
    "name":          _ADAPTOR_NAME,
    "capabilities":  _ADAPTOR_CAPABILITIES,
    "description":  """
The SGE (Sun/Oracle Grid Engine) adaptor allows to run and manage jobs on
`SGE <http://en.wikipedia.org/wiki/Oracle_Grid_Engine>`_ controlled HPC
clusters.
""",
    "example": "examples/jobs/sgejob.py",
    "schemas": {"sge":        "connect to a local cluster",
                "sge+ssh":    "conenct to a remote cluster via SSH",
                "sge+gsissh": "connect to a remote cluster via GSISSH"}
}

# --------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA
#
_ADAPTOR_INFO = {
    "name"         : _ADAPTOR_NAME,
    "version"      : "v0.1",
    "schemas"      : _ADAPTOR_SCHEMAS,
    "capabilities" : _ADAPTOR_CAPABILITIES,
    "cpis"         : [
                         {
                             "type": "radical.saga.job.Service",
                             "class": "SGEJobService"
                         },
                         {
                             "type": "radical.saga.job.Job",
                             "class": "SGEJob"
                         }
                     ]
}


###############################################################################
# The adaptor class
class Adaptor (a_base.Base):
    """ this is the actual adaptor class, which gets loaded by SAGA (i.e. by
        the SAGA engine), and which registers the CPI implementation classes
        which provide the adaptor's functionality.
    """

    # ----------------------------------------------------------------
    #
    def __init__(self):

        a_base.Base.__init__(self, _ADAPTOR_INFO)

        self.id_re = re.compile('^\[(.*)\]-\[(.*?)\]$')
        self.epoch = datetime(1970,1,1)

        self.purge_on_start   = self._cfg['purge_on_start']
        self.purge_older_than = self._cfg['purge_older_than']
        self.base_workdir     = self._cfg['base_workdir']


    # ----------------------------------------------------------------
    #
    def sanity_check(self):
        # FIXME: also check for gsissh
        pass


    # ----------------------------------------------------------------
    #
    def parse_id(self, id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(id)

        if not match or len(match.groups()) != 2:
            raise rse.BadParameter("Cannot parse job id '%s'" % id)

        return (match.group(1), match.group(2))


###############################################################################
#
class SGEJobService (cpi.Service):
    """ implements cpi.Service
    """

    # ----------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(SGEJobService, self)
        _cpi_base.__init__(api, adaptor)

        self._adaptor = adaptor

    # ----------------------------------------------------------------
    #
    def __del__(self):

        self.finalize(kill_shell=True)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        """ service instance constructor
        """

        self.rm      = rm_url
        self.session = session
        self.pe_list = list()
        self.jobs    = dict()
        self.queue   = None
        self.memreqs = None
        self.shell   = None
        self.mandatory_memreqs = list()
        self.accounting = False
        self.temp_path = self._adaptor.base_workdir


        rm_scheme = rm_url.scheme
        pty_url   = ru.Url (rm_url)

        # this adaptor supports options that can be passed via the
        # 'query' component of the job service URL.
        if rm_url.query is not None:
            for key, val in parse_qs(rm_url.query).items():
                if key == 'queue':
                    self.queue = val[0]
                elif key == 'memreqs':
                    self.memreqs = val[0]

        # we need to extrac the scheme for PTYShell. That's basically the
        # job.Serivce Url withou the sge+ part. We use the PTYShell to execute
        # pbs commands either locally or via gsissh or ssh.
        if rm_scheme == "sge":
            pty_url.scheme = "fork"
        elif rm_scheme == "sge+ssh":
            pty_url.scheme = "ssh"
        elif rm_scheme == "sge+gsissh":
            pty_url.scheme = "gsissh"

        # these are the commands that we need in order to interact with SGE.
        # the adaptor will try to find them during initialize(self) and bail
        # out in case they are not available.
        self._commands = {'qstat': None,
                          'qsub':  None,
                          'qdel':  None,
                          'qconf': None,
                          'qacct': None}

        self.shell = rsups.PTYShell(pty_url, self.session)

      # self.shell.set_initialize_hook(self.initialize)
      # self.shell.set_finalize_hook(self.finalize)

        self.initialize()

        return self.get_api ()


    # ----------------------------------------------------------------
    #
    def close (self) :
        if  self.shell :
            self.shell.finalize (True)

    # ----------------------------------------------------------------
    #
    def initialize(self):
        # check if all required sge tools are available
        for cmd in list(self._commands.keys()):
            ret, out, _ = self.shell.run_sync("which %s " % cmd)
            if ret != 0:
                raise rse.NoSuccess("Error finding SGE tools: %s" % out)
            else:
                path = out.strip()  # strip removes newline

                ret, out, _ = self.shell.run_sync("%s -help" % cmd)
                if ret != 0:
                    # fix for a bug in certain qstat versions that return
                    # '1' after a successfull qstat -help:
                    # https://github.com/radical-cybertools/radical.saga/issues/163
                    if cmd == 'qstat':
                        version = out.strip().split('\n')[0]
                    else:
                        raise rse.NoSuccess("Error finding SGE tools: %s" % out)
                else:
                    # version is reported in the first row of the
                    # help screen, e.g., GE 6.2u5_1
                    version = out.strip().split('\n')[0]

                    # add path and version to the command dictionary
                    self._commands[cmd] = {"path"   : "unset GREP_OPTIONS; %s"
                                                      % path,
                                           "version": version}

        self._logger.info("Found SGE tools: %s" % self._commands)

        # determine the available processing elements
        ret, out, _ = self.shell.run_sync('%s -spl' %
                      (self._commands['qconf']['path']))
        if ret != 0:
            raise rse.NoSuccess("Error running 'qconf': %s" % out)
        else:
            for pe in out.split('\n'):
                if pe != '':
                    self.pe_list.append(pe)
            self._logger.debug("Available processing elements: %s" %
                (self.pe_list))

        # find out mandatory and optional memory attributes
        ret, out, _ = self.shell.run_sync('%s -sc'
                                         % (self._commands['qconf']['path']))
        if ret != 0:
            raise rse.NoSuccess("Error running 'qconf': %s" % out)
        else:
            mandatory_attrs = []
            optional_attrs = []
            for line in out.split('\n'):
                if (line != '') and (line[0] != '#'):
                    [name, _, att_type, _, requestable, _, _, _] = line.split()
                    if att_type == 'MEMORY' and requestable == 'YES':
                        optional_attrs.append(name)
                    elif att_type == 'MEMORY' and requestable == 'FORCED':
                        mandatory_attrs.append(name)
            self._logger.debug("Optional  memory attr: %s" % optional_attrs)
            self._logger.debug("Mandatory memory attr: %s" % mandatory_attrs)

        # find out user specified memory attributes in job.Service URL
        if self.memreqs is None:
            flags = []
        else:
            flags, _ = self.__parse_memreqs(self.memreqs)

        # if there are mandatory memory attributes store them and check that
        # they were specified in the job.Service URL

        if not (mandatory_attrs == []):
            self.mandatory_memreqs = mandatory_attrs
            missing_flags = []
            for attr in mandatory_attrs:
                if attr not in flags:
                    missing_flags.append(attr)
            if missing_flags:
                raise rse.BadParameter("The following attribute(s) are "
                                       "mandatory in your SGE environment and "
                                       "thus must be specifiedin the job "
                                       "service URL: %s" % missing_flags)

        # if memory attributes were specified in the job.Service URL, check that
        # they correspond to existing optional or mandatory memory attributes

        invalid_attrs = []
        for f in flags:
            if not (f in optional_attrs or f in mandatory_attrs):
                invalid_attrs.append(f)

        if invalid_attrs:
            raise rse.BadParameter("The following attribute(s) are "
                                   "specifiedin the job service URL but "
                                   "are not accepted by your SGE "
                                   "environment: %s" % missing_flags)

        # check if accounting is activated
        qres = self.__kvcmd_results('qconf', '-sconf',
                                    filter_keys=["reporting_params"])
        self.accounting = "reporting_params" in qres and \
                          "accounting=true"  in qres["reporting_params"]

        # purge temporary files
        if self._adaptor.purge_on_start:
            cmd = "find " + self.temp_path + \
                  " -type f -mtime +%d -print -delete | wc -l" \
                  % self._adaptor.purge_older_than
            ret, out, _ = self.shell.run_sync(cmd)
            if ret == 0 and out != "0":
                self._logger.info("Purged %s temporary files" % out)


    # ----------------------------------------------------------------
    #
    def finalize(self, kill_shell=False):
        if  kill_shell :
            if  self.shell :
                self.shell.finalize (True)


    # ----------------------------------------------------------------
    #
    # private members
    #

    def __sge_to_saga_jobstate(self, sge_state):
        """
        Translates an SGE one-letter state to SAGA
        """

        try:
            if sge_state.startswith("d"):

                # when a qdel is done the state is prefixed with a d while the
                # termination signal is queued
                sge_state = sge_state[1:]

            return {
                'c'   : c.DONE,
                'E'   : c.RUNNING,
                'H'   : c.PENDING,
                'qw'  : c.PENDING,
                'r'   : c.RUNNING,
                't'   : c.RUNNING,
                'w'   : c.PENDING,
                's'   : c.PENDING,
                'X'   : c.CANCELED,
                'Eqw' : c.FAILED
            }[sge_state]
        except:
            return c.UNKNOWN

    def __parse_memreqs(self, s):
        """
        Simple parser for getting memory requirements flags and multipliers
        from the memreqs part of the job.Service url
        """
        flags = []
        multipliers = []
        while len(s) != 0:
            # find multiplier
            m = re.match(r'\d+\.?\d*|\d*\.?\d+', s)
            if m:
                multipliers.append(float(s[m.start():m.end()]))
                s = s[m.end():]
            else:
                multipliers.append(1.0)
            # find flag
            pos = s.find('~')
            if pos < 0:
                flags.append(s)
                s = ''
            else:
                flags.append(s[:pos])
                s = s[pos + 1:]
        return flags, multipliers


    def __kvcmd_results(self, cmd, cmd_args, *args, **kwargs):
        """
        Runs a SGE command that returns key-value pairs as result and parses
        the results.

        :param cmd: command alias
        :param cmd_args: command arguments
        :param args: parser arguments
        :param kwargs: parser keyword arguments
        :returns: a dictionary if succeeded or None otherwise
        """

        ret, out, _ = self.shell.run_sync(
                              '%s %s' % (self._commands[cmd]['path'], cmd_args))
        if ret == 0:
            return SgeKeyValueParser(out, *args, **kwargs).as_dict()
        return None

    def __remote_mkdir(self, path):
        """
        Creates a directory on the remote host.
        :param path: the remote directory to be created.
        """
        # check if the path exists
        ret, out, _ = self.shell.run_sync(
                       "(test -d %s && echo -n 0) || (mkdir -p %s && echo -n 1)"
                       % (path, path))

        if ret == 0 and out == "1":
            self._logger.info("Remote directory created: %s" % path)

        elif ret != 0:
            # something went wrong
            raise rse.NoSuccess("Couldn't create remote directory - %s" % out)

    def __job_info_from_accounting(self, sge_job_id, max_retries=10):
        """
        Returns job information from the SGE accounting using qacct.

        It may happen that when the job exits from the queue system the results
        in the accounting database take some time to appear. To avoid premature
        failing several tries can be done (up to a maximum) with delays of
        1 second in between.

        :param sge_job_id: SGE job id
        :param max_retries: The maximum number of retries in case qacct fails
        :return: job information dictionary
        """

        job_info = None
        retries  = max_retries

        while retries > 0:
            retries -= 1

            qres = self.__kvcmd_results('qacct', "-j %s | grep -E '%s'"
                                       % (sge_job_id,
           "jobname|hostname|qsub_time|start_time|end_time|exit_status|failed"))

            if not qres and retries > 0:
                # sometimes there is a lapse between the job exits from the
                # queue and its information enters in the accounting database
                # let's run qacct again after a delay
                time.sleep(1)
                continue


            # ok, extract job info from qres
            # jobname      test
            # hostname     sge
            # qsub_time    Mon Jun 24 17:24:43 2013  # FIXME: convert to EPOCH
            # start_time   Mon Jun 24 17:24:50 2013  # FIXME: convert to EPOCH
            # end_time     Mon Jun 24 17:44:50 2013  # FIXME: convert to EPOCH
            # failed       0
            # exit_status  0

            if qres.get("failed") == "0": state = c.DONE
            else                        : state = c.FAILED
            job_info = {'state'       : state,
                        'name'        : qres.get("jobname"),
                        'exec_hosts'  : qres.get("hostname"),
                        'create_time' : qres.get("qsub_time"),
                        'start_time'  : qres.get("start_time"),
                        'end_time'    : qres.get("end_time"),
                        'returncode'  : int(qres.get("exit_status", -1)),
                        'gone'        : False
                       }
            break

        return job_info


    def __remote_job_info_path(self, sge_job_id="$JOB_ID"):
        """
        Returns the path of the remote job info file.
        param sge_job_id: the SGE job id, if omitted an enviroment variable
                           representing the job id will be used.
        return: path to the remote job info file
        """

        return "%s/%s" % (self.temp_path, sge_job_id)


    def __clean_remote_job_info(self, sge_job_id):
        """
        Removes the temporary remote file containing job info.
        :param sge_job_id: the SGE job id
        """

        path = self.__remote_job_info_path(sge_job_id)
        ret, out, _ = self.shell.run_sync("rm %s" % path)
        if ret != 0:
            self._logger.debug("Remote job info couldn't be removed: %s" % path)


    def __get_remote_job_info(self, sge_job_id):
        """
        Obtains job info from temporary remote file created by the qsub script.
        :param sge_job_id: the SGE job id
        :return: a dictionary with the job info
        """

        ret, out, _ = self.shell.run_sync("cat %s" % self.__remote_job_info_path(sge_job_id))
        if ret != 0:
            return None

        qres = SgeKeyValueParser(out, key_suffix=":").as_dict()

        if   "signal"      in qres: state = c.CANCELED
        elif "exit_status" in qres: state = c.DONE
        else                      : state = c.RUNNING

        job_info = {'state'       : state,
                    'name'        : qres.get("jobname"),
                    'exec_hosts'  : qres.get("hostname"),
                    'create_time' : qres.get("qsub_time"),
                    'start_time'  : qres.get("start_time"),
                    'end_time'    : qres.get("end_time"),
                    'returncode'  : int(qres.get("exit_status", -1)),
                    'gone'        : False}

        return job_info

    def __generate_qsub_script(self, jd):
        """
        Generates an SGE script from a SAGA job description
        :param jd: job descriptor
        :return: the qsub script
        """

        # SGE parameters

        sge_params = ["#$ -S /bin/bash"]

        if jd.name is not None:
            sge_params += ["#$ -N %s" % jd.name]

        sge_params += ["#$ -V"]

        if jd.environment is not None and len(jd.environment) > 0:
            env_list = ",".join(["%s=%s" % (key, value) for key, value
                                                     in list(jd.environment.items())])
            sge_params += ["#$ -v %s" % env_list]

        if jd.working_directory is not None:
            sge_params += ["#$ -wd %s" % jd.working_directory]

        if jd.output is not None:
            sge_params += ["#$ -o %s" % jd.output]

        if jd.error is not None:
            sge_params += ["#$ -e %s" % jd.error]

        if jd.wall_time_limit is not None:
            hours = int(jd.wall_time_limit / 60)
            minutes = jd.wall_time_limit % 60
            sge_params += ["#$ -l h_rt=%s:%s:00" % (str(hours), str(minutes))]

        queue = self.queue or jd.queue
        if queue is not None:
            sge_params += ["#$ -q %s" % queue]

        if jd.project is not None:
            sge_params += ["#$ -A %s" % str(jd.project)]

        if jd.job_contact is not None:
            sge_params += ["#$ -m be", "#$ -M %s" % jd.contact]

        # memory requirements - TOTAL_PHYSICAL_MEMORY
        # it is assumed that the value passed through jd is always in Megabyte
        if jd.total_physical_memory is not None:
            # this is (of course) not the same for all SGE installations. some
            # use virtual_free, some use a combination of mem_req / h_vmem.
            # It is very annoying. We need some sort of configuration variable
            # that can control this. Yes, ugly and not very saga-ish, but
            # the only way to do this, IMHO...
            if self.memreqs is None:
                raise Exception("""
When using 'total_physical_memory' with the SGE adaptor, the query parameters of
the job.Service URL must define the attributes used by your particular instance
of SGE to control memory allocation.  'virtual_free', 'h_vmem' or 'mem_req' are
commonly encountered examples of such attributes.
A valid job.Service URL could be for instance:

    'sge+ssh://myserver.edu?memreqs=virtual_free~1.5h_vmem'

here the attribute 'virtual_free' would be set to 'total_physical_memory' and
the attribute 'h_vmem' would be set to 1.5*'total_physical_memory' '~' is used
as a separator.
""")

            flags, multipliers = self.__parse_memreqs(self.memreqs)
            for flag, mult in zip(flags, multipliers):
                sge_params += ["#$ -l %s=%sm" % (flag,
                            int(round(mult * int(jd.total_physical_memory))))]

        # check spmd variation. this translates to the SGE qsub -pe flag.
        if jd.spmd_variation is not None:
            if jd.spmd_variation not in self.pe_list:
                raise Exception("'%s' is not valid for jd.spmd_variation. "
                      "Valid are: %s" % (jd.spmd_variation, self.pe_list))

            # if no cores are requested at all, we default to 1

            # we need to translate the # cores requested into
            # multiplicity, i.e., if one core is requested and
            # the cluster consists of 16-way SMP nodes, we will
            # request 16. If 17 cores are requested, we will
            # request 32... and so on ... self.__ppn represents
            # the core count per single node
            #count = int(int(jd.total_cpu_count) / int(ppn))
            #if int(jd.total_cpu_count) % int(ppn) != 0:
            #    count = count + 1
            #count = count * int(ppn)
            sge_params += ["#$ -pe %s %s" % (jd.spmd_variation,
                                             jd.total_cpu_count or 1)]

        elif jd.total_cpu_count is not None and jd.total_cpu_count > 1:
                raise Exception("jd.total_cpu_count requires that "
                                "jd.spmd_variation is not empty. "
                                "Valid are: %s" % (self.pe_list))


        # CANDEIDATE_HOSTS - this translates to 'qsub -l h="host1|host2|..."'
        if jd.candidate_hosts:
            sge_params.append('#$ -l h="%s"' % '|'.join(jd.candidate_hosts))


        # convert sge params into an string
        sge_params = "\n".join(sge_params)

        # Job info, executable and arguments

        job_info_path = self.__remote_job_info_path()

        script_body = [
            'function aborted() {',
            '  echo Aborted with signal $1.',
            '  echo "signal: $1" >>%s' % job_info_path,
            '  echo "end_time: $(LC_ALL=en_US.utf8 date \'+%%s\')" >>%s'
                    % job_info_path,
            '  exit -1',
            '}',
            'mkdir -p %s' % self.temp_path,
            'for sig in SIGHUP SIGINT SIGQUIT SIGTERM SIGUSR1 SIGUSR2;'
            '    do trap "aborted $sig" $sig; done',
            'echo "hostname: $HOSTNAME" >%s' % job_info_path,
            'echo "jobname: %s" >>%s' % (jd.name, job_info_path),
            'echo "qsub_time: %s" >>%s' % (time.time(), job_info_path),
            'echo "start_time: $(LC_ALL=en_US.utf8 date \'+%%s\')" >>%s'
                   % job_info_path
        ]

        exec_n_args = None
        if jd.executable is not None:
            exec_n_args = jd.executable
            if jd.arguments is not None:
                exec_n_args += " %s" % " ".join(jd.arguments)

        elif jd.arguments is not None:
            raise Exception("jd.arguments defined without jd.executable")

        if exec_n_args is not None:
            script_body += [exec_n_args]

        script_body += [
            'echo "exit_status: $?" >>%s' % job_info_path,
            'echo "end_time: $(LC_ALL=en_US.utf8 date \'+%%s\')" >>%s'
                  % job_info_path
        ]

        # convert exec and args into an string and
        # escape all double quotes and dollar signs, otherwise 'echo |'
        # further down won't work.
        # only escape '$' in args and exe. not in the params
        script_body = "\n".join(script_body).replace('$', '\\$')

        sgescript = "\n#!/bin/bash \n%s \n%s" % (sge_params, script_body)

        return sgescript.replace('"', '\\"')

    # ----------------------------------------------------------------
    #
    # Adaptor internal methods
    #

    def _job_run(self, jd):
        """
        Runs a job via qsub
        """

        if self.queue and jd.queue and self.queue != jd.queue:
            self._logger.warning("Job service has explicit 'queue=%s', "
                                 "but job description uses queue: '%s'. "
                                 "Using '%s'."
                                 % (self.queue, jd.queue, self.queue))

        # In SGE environments with mandatory memory attributes,
        # 'total_physical_memory' must be specified
        if self.mandatory_memreqs and jd.total_physical_memory is None:
            raise rse.BadParameter("Your SGE environments has mandatory "
                                "attributes, so 'total_physical_memory' "
                                "must be specified in your job descriptor")

        try:
            # create a SGE job script from SAGA job description
            script = self.__generate_qsub_script(jd)
            self._logger.info("Generated SGE script: %s" % script)

        except Exception as e:
            raise rse.BadParameter('error generating SGE script') from e

        # try to create the working/output/error directories (if defined)
        # WARNING: this assumes a shared filesystem between login node and
        #           compute nodes.
        if jd.working_directory is not None and len(jd.working_directory) > 0:
            self.__remote_mkdir(jd.working_directory)

        if jd.output is not None and len(jd.output) > 0:
            self.__remote_mkdir(os.path.dirname(jd.output))

        if jd.error is not None and len(jd.error) > 0:
            self.__remote_mkdir(os.path.dirname(jd.output))

        # submit the SGE script
        # Now we want to execute the script. This process consists of two steps:
        # (1) we create a temporary file with 'mktemp' and write the contents of
        #     the generated PBS script into it
        # (2) we call 'qsub <tmpfile>' to submit the script to the queueing
        #     system
        cmdline = 'SCRIPTFILE=`mktemp -t RS-SGEJobScript.XXXXXX` ' \
                  ' && echo "%s" > $SCRIPTFILE ' \
                  ' && %s -notify $SCRIPTFILE ' \
                  ' && rm -f $SCRIPTFILE' \
                  %  (script, self._commands['qsub']['path'])

        ret, out, _ = self.shell.run_sync(cmdline)

        if ret != 0:
            # something went wrong
            raise rse.NoSuccess("Error running qsub: %s [%s]" % (out, cmdline))

        # stdout contains the job id:
        # Your job 1036608 ("testjob") has been submitted
        sge_job_id = None
        for line in out.split('\n'):
            if line.find("Your job") != -1:
                sge_job_id = line.split()[2]
        if sge_job_id is None:
            raise rse.NoSuccess("Couldn't parse job id from 'qsub': %s" % out)

        job_id = "[%s]-[%s]" % (self.rm, sge_job_id)
        self._logger.info("Submitted SGE job with id: %s" % job_id)

        # add job to internal list of known jobs.
        self.jobs[job_id] = {
            'state':        c.PENDING,
            'name':         jd.name,
            'exec_hosts':   None,
            'returncode':   None,
            'create_time':  None,
            'start_time':   None,
            'end_time':     None,
            'gone':         False
        }

        return job_id

    # ----------------------------------------------------------------
    #
    def _retrieve_job(self, job_id):
        """ retrieve job information
        :param job_id: SAGA job id
        :return: job information dictionary
        """

        rm, pid = self._adaptor.parse_id(job_id)

        # check the state of the job
        ret, out, _ = self.shell.run_sync(
                    "%s | tail -n+3 | awk '($1==%s) {{print $5,$6,$7,$8}}'" % (
                    self._commands['qstat']['path'], pid))

        out = out.strip()

        job_info = None

        if ret == 0 and len(out) > 0:

            # job is still in the queue
            # output is something like
            # r 06/24/2013 17:24:50

            m = _QSTAT_JOB_STATE_RE.match(out)
            if not m:
                # something wrong with the result of qstat
                raise rse.NoSuccess("Unexpected qstat results:\n%s" % out.rstrip())

            state, start_time, queue = m.groups()

            # Convert start time into EPOCH
            start_time = None
            if state in ["r", "t", "s", "S", "T", "d", "E", "Eqw"]:
                try:
                    dt = datetime.strptime(start_time, "%m/%d/%Y %H:%M:%S")
                    start_time = (dt - self._adaptor.epoch).total_seconds()
                except:
                    # keep 'None'
                    pass

            exec_host = None
            if "@" in queue:
                queue, exec_host = queue.split("@")
                exec_host = exec_host.rstrip()

            # if it is an Eqw job it is better to retrieve the information
            # from qacct
            if self.accounting and state == "Eqw":
                job_info = self.__job_info_from_accounting(pid)
                # TODO remove the job from the queue ?

            if job_info is None: # use qstat -j pid
                qres = self.__kvcmd_results(
                'qstat', "-j %s | grep -E 'job_name|submission_time|sge_o_host'"
                % pid, key_suffix=":")

                if qres is not None:

                    # when qstat fails it will fall back to qacct
                    # output is something like
                    # submission_time:            Mon Jun 24 17:24:43 2013
                    # sge_o_host:                 sge

                    state = self.__sge_to_saga_jobstate(state)
                    host  = exec_host or qres.get("sge_o_host")
                    job_info = {'state'       : state,
                                'name'        : qres.get("job_name"),
                                'exec_hosts'  : host,
                                'returncode'  : None,
                                'create_time' : qres.get("submission_time"),
                                'start_time'  : start_time,
                                'end_time'    : None,
                                'gone'        : False
                                }

        # if job already finished or there was an error with qstat
        # try to read the remote job info
        if job_info is None:
            job_info = self.__get_remote_job_info(pid)

        # none of the previous methods gave us job info
        # if accounting is activated use qacct
        if self.accounting and job_info is None:
            job_info = self.__job_info_from_accounting(pid)

        if job_info is None: # Oooops, we couldn't retrieve information from SGE
            raise rse.NoSuccess("Couldn't reconnect to job '%s'" % job_id)

        elems = ["name", "state", "returncode", "exec_hosts", "create_time",
                 "start_time", "end_time", "gone"]
        self._logger.debug("job_info(%s)=[%s]"
                % (pid, ", ".join(["%s=%s" % (k, str(job_info[k]))
                                           for k in elems])))
        return job_info


    # ----------------------------------------------------------------
    #
    def _job_get_info(self, job_id):
        """ get job attributes
        """

        # if we don't have the job in our dictionary, we don't want it
        if job_id not in self.jobs:
            raise rse.NoSuccess("Unknown job ID: %s" % job_id)

        # prev. info contains the info collect when _job_get_info
        # was called the last time
        prev_info = self.jobs[job_id]

        # if the 'gone' flag is set, there's no need to query the job
        # state again. it's gone forever
        if prev_info['gone'] is True:
            self._logger.warning("Job information is not available anymore.")
            return prev_info

        # if the job is in a terminal state don't expect it to change anymore
        if prev_info["state"] in c.FINAL:
            return prev_info

        # retrieve updated job information
        curr_info = self._retrieve_job(job_id)
        if curr_info is None:
            prev_info["gone"] = True
            return prev_info

        # update the job info cache and return it
        self.jobs[job_id] = curr_info
        return curr_info


    # ----------------------------------------------------------------
    #
    def _job_get_state(self, job_id):
        """ get the job's state
        """
        # check if we have already reach a terminal state
        if self.jobs[job_id]['state'] in c.FINAL:
            return self.jobs[job_id]['state']

        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['state']


    # ----------------------------------------------------------------
    #
    def _job_get_exit_code(self, job_id):
        """ get the job's exit code
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['returncode'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        ret = self.jobs[job_id]['returncode']

        # FIXME: 'None' should cause an exception
        if ret == None : return None
        else           : return int(ret)


    # ----------------------------------------------------------------
    #
    def _job_get_execution_hosts(self, job_id):
        """ get the job's exit code
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['exec_hosts'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['exec_hosts']


    # ----------------------------------------------------------------
    #
    def _job_get_create_time(self, job_id):
        """ get the job's creation time
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['create_time'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        # FIXME: convert to EOPCH
        return self.jobs[job_id]['create_time']


    # ----------------------------------------------------------------
    #
    def _job_get_start_time(self, job_id):
        """ get the job's start time
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['start_time'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['start_time']


    # ----------------------------------------------------------------
    #
    def _job_get_end_time(self, job_id):
        """ get the job's end time
        """
        # check if we can / should update
        if (self.jobs[job_id]['gone'] is not True) \
        and (self.jobs[job_id]['end_time'] is None):
            self.jobs[job_id] = self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['end_time']


    # ----------------------------------------------------------------
    #
    def _job_cancel(self, job_id):
        """ cancel the job via 'qdel'
        """
        rm, pid = self._adaptor.parse_id(job_id)

        ret, out, _ = self.shell.run_sync("%s %s\n" \
            % (self._commands['qdel']['path'], pid))

        if ret != 0:
            raise rse.NoSuccess("Error canceling job via 'qdel': %s" % out)

        self.__clean_remote_job_info(pid)

        # assume the job was succesfully canceld
        self.jobs[job_id]['state'] = c.CANCELED


    # ----------------------------------------------------------------
    #
    def _job_wait(self, job_id, timeout):
        """ wait for the job to finish or fail
        """

        time_start = time.time()
        time_now   = time_start
        rm, pid    = self._adaptor.parse_id(job_id)

        while True:
            state = self._job_get_state(job_id=job_id)

            if state == c.UNKNOWN :
                raise rse.IncorrectState("cannot get job state")

            if state in c.FINAL:
                self.__clean_remote_job_info(pid)
                return True
            # avoid busy poll
            time.sleep(0.5)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        """ implements cpi.Service.get_url()
        """
        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         "job_description": jd,
                         "job_schema":      self.rm.schema,
                         "reconnect":       False
                        }

        return api.Job(_adaptor=self._adaptor,
                       _adaptor_state=adaptor_state)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, jobid):
        """ Implements cpi.Service.get_job()
        """

        # try to get some information about this job
        job_info = self._retrieve_job(jobid)

        # save it into our job dictionary.
        self.jobs[jobid] = job_info

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         # TODO: fill job description
                         "job_description": api.Description(),
                         "job_schema":      self.rm.schema,
                         "reconnect":       True,
                         "reconnect_jobid": jobid
                        }

        return api.Job(_adaptor=self._adaptor,
                       _adaptor_state=adaptor_state)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        """ implements cpi.Service.get_url()
        """
        return self.rm


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        """ implements cpi.Service.list()
        """
        ids = []

        ret, out, _ = self.shell.run_sync(
                           "unset GREP_OPTIONS; %s | grep `whoami`"
                           % self._commands['qstat']['path'])

        if ret != 0 and len(out) > 0:
            raise rse.NoSuccess("Failed to list jobs via 'qstat': %s" % out)

        elif ret != 0 and len(out) == 0:
            # qstat | grep `whoami` exits with 1 if the list is empty
            pass

        else:
            for line in out.split("\n"):
                if len(line.split()) > 1:
                    jobid = "[%s]-[%s]" % (self.rm, line.split()[0])
                    ids.append(jobid)

        return ids


  # # ----------------------------------------------------------------
  # #
  # def container_run (self, jobs) :
  #     self._logger.debug ("container run: %s"  %  str(jobs))
  #     # TODO: this is not optimized yet
  #     for job in jobs:
  #         job.run ()
  #
  #
  # # ----------------------------------------------------------------
  # #
  # def container_wait (self, jobs, mode, timeout) :
  #     self._logger.debug ("container wait: %s"  %  str(jobs))
  #     # TODO: this is not optimized yet
  #     for job in jobs:
  #         job.wait ()
  #
  #
  # # ----------------------------------------------------------------
  # #
  # def container_cancel (self, jobs, timeout) :
  #     self._logger.debug ("container cancel: %s"  %  str(jobs))
  #     raise rse.NoSuccess ("Not Implemented");


###############################################################################
#
class SGEJob (cpi.Job):
    """ implements cpi.Job
    """

    def __init__(self, api, adaptor):

        # initialize parent class
        _cpi_base = super(SGEJob, self)
        _cpi_base.__init__(api, adaptor)

    @SYNC_CALL
    def init_instance(self, job_info):
        """ implements cpi.Job.init_instance()
        """
        # init_instance is called for every new api.Job object
        # that is created
        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        if job_info['reconnect'] is True:
            self._id      = job_info['reconnect_jobid']
            self._name    = self.jd.name
            self._started = True
        else:
            self._id      = None
            self._name    = self.jd.name
            self._started = False

        return self.get_api()

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description (self):
        return self.jd

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        """ implements cpi.Job.get_state()
        """
        if self._started is False:
            # jobs that are not started are always in 'NEW' state
            return api.NEW
        else:
            return self.js._job_get_state(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):
        """ implements cpi.Job.wait()
        """
        if self._started is False:
            raise rse.IncorrectState("Can't wait for job that hasn't been started")
        else:
            self.js._job_wait(self._id, timeout)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):
        """ implements cpi.Job.cancel()
        """
        if self._started is False:
            raise rse.IncorrectState("Can't wait for job that hasn't been started")
        else:
            self.js._job_cancel(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):
        """ implements cpi.Job.run()
        """
        self._id = self.js._job_run(self.jd)
        self._started = True

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):
        """ implements cpi.Job.get_service_url()
        """
        return self.js.rm

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):
        """ implements cpi.Job.get_id()
        """
        return self._id

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name (self):
        """ Implements cpi.Job.get_name() """
        return self._name

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):
        """ implements cpi.Job.get_exit_code()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_exit_code(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):
        """ implements cpi.Job.get_created()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_create_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):
        """ implements cpi.Job.get_started()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_start_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):
        """ implements cpi.Job.get_finished()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_end_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):
        """ implements cpi.Job.get_execution_hosts()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_execution_hosts(self._id)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner, Mark Santcroos, Manuel Maldonado"
__copyright__ = "Copyright 2012-2016, The SAGA Project"
__license__   = "MIT"


""" Cobalt job adaptor implementation
"""

import re
import os
import time
import datetime
import threading

from urllib.parse import parse_qs

from ...url           import Url
from ...              import exceptions as rse
from ...utils         import pty_shell  as rsups
from ...              import job        as api
from ..               import base       as a_base
from ..cpi            import job        as cpi_job
from ..cpi            import decorators as cpi_decs


SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL

SYNC_WAIT_UPDATE_INTERVAL =  1  # seconds
MONITOR_UPDATE_INTERVAL   = 60  # seconds


# ------------------------------------------------------------------------------
#
class _job_state_monitor(threading.Thread):
    """ thread that periodically monitors job states
    """
    def __init__(self, job_service):

        self.logger = job_service._logger
        self.js = job_service
        self._term = threading.Event()

        super(_job_state_monitor, self).__init__()
        self.setDaemon(True)

    def stop(self):
        self._term.set()


    def run(self):

        # we stop the monitoring thread when we see the same error 3 times in
        # a row...
        error_type_count = dict()

        while not self._term.is_set ():

            try:
                # FIXME: do bulk updates here! we don't want to pull information
                # job by job. that would be too inefficient!
                jobs = self.js.jobs

                for job_id in list(jobs.keys()) :

                    job_info = jobs[job_id]

                    # we only need to monitor jobs that are not in a
                    # terminal state, so we can skip the ones that are
                    # either done, failed or canceled
                    if  job_info['state'] \
                        not in [api.DONE, api.FAILED, api.CANCELED]:

                        # Store the current state since the current state
                        # variable is updated when _job_get_info is called
                        pre_update_state = job_info['state']

                        new_job_info = self.js._job_get_info(job_id, reconnect=False)
                        self.logger.info ("Job monitoring thread updating Job "
                                          "%s (old state: %s, new state: %s)" %
                                          (job_id, pre_update_state, new_job_info['state']))

                        # fire job state callback if 'state' has changed
                        if  new_job_info['state'] != pre_update_state:
                            job_obj = job_info['obj']
                            job_obj._attributes_i_set('state', new_job_info['state'], job_obj._UP, True)

                        # update job info
                        jobs[job_id] = new_job_info

            except Exception as e:
                import traceback
                traceback.print_exc()
                self.logger.warning("Exception caught in job monitoring thread: %s" % e)

                # check if we see the same error again and again
                error_type = str(e)
                if  error_type not in error_type_count :
                    error_type_count = dict()
                    error_type_count[error_type]  = 1
                else :
                    error_type_count[error_type] += 1
                    if  error_type_count[error_type] >= 3 :
                        self.logger.error("too many monitoring errors -- stopping job monitoring thread")
                        return

            finally :
                time.sleep (MONITOR_UPDATE_INTERVAL)


# ------------------------------------------------------------------------------
#
def _cobalt_to_saga_jobstate(cobaltjs):
    """ translates a cobalt one-letter state to saga
    """

    if   cobaltjs == 'C': return api.DONE
    elif cobaltjs == 'F': return api.DONE
    elif cobaltjs == 'H': return api.PENDING   # held by user or dependency
    elif cobaltjs == 'Q': return api.PENDING   # queued
    elif cobaltjs == 'S': return api.PENDING   # suspended
    elif cobaltjs == 'W': return api.PENDING   # waiting for execution
    elif cobaltjs == 'R': return api.RUNNING   # starting/running
    elif cobaltjs == 'E': return api.RUNNING   # exiting after having run
    elif cobaltjs == 'T': return api.RUNNING   # being moved to new location
    elif cobaltjs == 'X': return api.CANCELED  # Subjob completed or deleted
    else                : return api.UNKNOWN


# ------------------------------------------------------------------------------
#
def _cobaltscript_generator(url, logger, jd, ppn, is_cray=False, queue=None,
        run_job=None):
    """
    Generates Cobalt-style 'qsub' command arguments from a SAGA job description
    """
    cobalt_params       = str()
    exec_n_args         = str()
    cobaltscript        = str()
    total_cpu_count     = 1  # default value
    number_of_processes = None
    processes_per_host  = None
    blue_gene_q_modes   = [1, 2, 4, 8, 16, 32, 64]

    if jd.executable:
        exec_n_args += "%s " % (jd.executable)
    if jd.arguments:
        for arg in jd.arguments:
            exec_n_args += "%s " % (arg)

    if jd.name:
        cobalt_params += '#COBALT --jobname %s\n' % jd.name

    if jd.working_directory:
        if not os.path.isabs(jd.working_directory):
            jd.working_directory = '$HOME/%s' % jd.working_directory
        cobalt_params += '#COBALT --cwd %s\n' % jd.working_directory

    # a workaround is to do an explicit 'cd'
    # if jd.working_directory:
    #     workdir_directives  = 'export    PBS_O_WORKDIR=%s \n' \
    #                                            % jd.working_directory
    #     workdir_directives += 'mkdir -p  %s\n' % jd.working_directory
    #     workdir_directives += 'cd        %s\n' % jd.working_directory
    # else:
    #     workdir_directives = ''

    if jd.output:
        # if working directory is set, we want stdout to end up in
        # the working directory as well, unless it containes a specific
        # path name.
        if jd.working_directory:
            if os.path.isabs(jd.output):
                cobalt_params += '#COBALT --output %s\n' % jd.output
            else:
                # user provided a relative path for STDOUT. in this case
                # we prepend the workind directory path before passing
                # it on to Cobalt
                cobalt_params += '#COBALT --output %s/%s\n' \
                               % (jd.working_directory, jd.output)
        else:
            cobalt_params += '#COBALT --output %s\n' % jd.output

    if jd.error:
        # if working directory is set, we want stderr to end up in
        # the working directory as well, unless it contains a specific
        # path name.
        if jd.working_directory:
            if os.path.isabs(jd.error):
                cobalt_params += '#COBALT --error %s\n' % jd.error
            else:
                # user provided a realtive path for STDERR. in this case
                # we prepend the workind directory path before passing
                # it on to Cobalt
                cobalt_params += '#COBALT --error %s/%s\n' \
                               % (jd.working_directory, jd.error)
        else:
            cobalt_params += '#COBALT --error %s\n' % jd.error

    if jd.wall_time_limit:
        hours = int(jd.wall_time_limit / 60)
        minutes = jd.wall_time_limit % 60
        cobalt_params += '#COBALT --time %s:%s:00\n' \
            % (str(hours).zfill(2), str(minutes).zfill(2))

    if queue:
        cobalt_params += '#COBALT --queue %s\n' % queue
    elif jd.queue:
        cobalt_params += '#COBALT --queue %s\n' % jd.queue

    if jd.project:
        cobalt_params += '#COBALT --project %s\n' % str(jd.project)

    if jd.job_contact:
        cobalt_params += '#COBALT --notify %s\n' % str(jd.job_contact)

    #
    # This section takes care of CPU/Process/Node calculation
    #

    # handle number of cores
    if jd.attribute_exists('total_cpu_count'):
        total_cpu_count = jd.total_cpu_count

    # ppn: processes/cores per node (default value is an input parameter)
    if jd.attribute_exists('processes_per_host'):
        ppn = jd.processes_per_host

    # Request enough nodes to cater for the number of cores requested
    number_of_nodes = total_cpu_count / ppn
    if total_cpu_count % ppn > 0:
        number_of_nodes += 1

    # Get number of processes
    # Defaults to number_of_processes = number_of_nodes
    if jd.attribute_exists('number_of_processes'):
        number_of_processes = jd.number_of_processes
    else:
        logger.debug("number_of_processes not specified. default: 1 per node")
        number_of_processes = number_of_nodes

  # # Get number of processes per host/node
  # # Defaults to processes_per_host = 1
  # if jd.attribute_exists("processes_per_host"):
  #     processes_per_host = jd.processes_per_host
  # else:
  #     logger.debug("processes_per_host not specified -- default to 1")
  #     processes_per_host = 1
  #
  # # Need to make sure that the 'processes_per_host' is a valid one
  # # Blue Gene/Q valid modes ==> [1, 2, 4, 8, 16, 32, 64]
  # # At the Blue Gene/Q, 1 Node == 16 Cores
  # # and can handle UP TO 4 tasks per CPU/Core == 64 tasks
  # #
  # # References:
  # #   http://www.alcf.anl.gov/user-guides/cobalt-job-control
  # #   https://www.alcf.anl.gov/user-guides/blue-geneq-versus-blue-genep
  # if processes_per_host not in blue_gene_q_modes:
  #     log_error_and_raise("#processes per host %d incompatible with #nodes %d"
  #                        % (processes_per_host, blue_gene_q_modes),
  #                           rse.BadParameter, logger)
  #
  # # Make sure we aren't doing funky math
  # # References:
  # #   http://www.alcf.anl.gov/user-guides/machine-partitions
  # #   the --proccount flag value must be <= nodecount * mode
  # if  number_of_processes > (number_of_nodes * processes_per_host):
  #     log_error_and_raise (("number_of_processes (%d) must be <= to"
  #         "(number_of_nodes * processes_per_host) (%d * %d = %d)")
  #         % (number_of_processes, number_of_nodes, processes_per_host,
  #             (number_of_nodes * processes_per_host)), rse.NoSuccess, logger)

    # Other funky math checks should go here ~

    # Set number of nodes
    cobalt_params += '#COBALT --nodecount %d\n' % number_of_nodes

    # Set the total number of processes
    cobalt_params += '#COBALT --proccount %d\n' % number_of_processes

    # The Environments are added at the end because for now
    # Cobalt isn't supporting spaces in the env variables...
    # Which mess up the whole script if they are at the begining of the list...
    if jd.environment:
        cobalt_params += '#COBALT --env %s\n' % \
                ':'.join (["%s=%s" % (k,v.replace(':', '\\:')
                                         .replace('=', '\\='))  # escape chars
                           for k,v in jd.environment.items()])

    # Why do I need this?
    # Andre on Dev 16 2016:
    # You don't, but it can be useful for some applications,
    # so this is supposed to be available in the job environment for inspection.
    # This makes sense to be exported as an environment Variable
    cobalt_params += '#COBALT --env SAGA_PPN=%d\n' % ppn

    # may not need to escape all double quotes and dollarsigns,
    # since we don't do 'echo |' further down (like torque/pbspro)
    # only escape '$' in args and exe. not in the params
    # exec_n_args = exec_n_args.replace('$', '\\$')

    # Set the MPI rank per node (mode).
    #   mode --> c1, c2, c4, c8, c16, c32, c64
    #   Mode is represented by the runjob's '--ranks-per-node' flag
    exec_n_args = exec_n_args.replace('$', '\\$')

    # Need a new line before the shebang because linux is a bit of a pain when
    # echoing it.  It will be removed later though...
    cobaltscript = "\n#!/bin/bash\n%s\n%s\n\n" % (cobalt_params, exec_n_args)
    cobaltscript = cobaltscript.replace('"', '\\"')
    return cobaltscript


# ------------------------------------------------------------------------------
# some private defs
#
_PTY_TIMEOUT = 2.0

# ------------------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.cobaltjob"
_ADAPTOR_SCHEMAS       = ["cobalt", "cobalt+ssh", "cobalt+gsissh"]
_ADAPTOR_OPTIONS       = [
    {
        'category'         : 'radical.saga.adaptors.cobaltjob',
        'name'             : 'base_workdir',
        'type'             : str,
        'default'          : "$HOME/.radical/saga/adaptors/cobaltjob/",
        'documentation'    : '''The adaptor stores job state information on the
                              filesystem on the target resource. This parameter
                              specified what location should be used.''',
        'env_variable'     : None
    },
    {
        'category'         : 'radical.saga.adaptors.cobaltjob',
        'name'             : 'purge_on_start',
        'type'             : bool,
        'default'          : True,
        'valid_options'    : [True, False],
        'documentation'    : '''Purge temporary job information for all
                              jobs which are older than a number of days.
                              The number of days can be configured with
                              <purge_older_than>.''',
        'env_variable'     : None
    },
    {
        'category'         : 'radical.saga.adaptors.cobaltjob',
        'name'             : 'purge_older_than',
        'type'             : int,
        'default'          : 30,
        'documentation'    : '''When <purge_on_start> is enabled this specifies
                                the number of days to consider a temporary file
                                older enough to be deleted.''',
        'env_variable'     : None
    },
]

# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES = {
    "jdes_attributes":   [api.NAME,
                          api.EXECUTABLE,
                          api.ARGUMENTS,
                          api.ENVIRONMENT,
                          api.INPUT,
                          api.OUTPUT,
                          api.ERROR,
                          api.QUEUE,
                          api.PROJECT,
                          api.WALL_TIME_LIMIT,
                          api.WORKING_DIRECTORY,
                          api.SPMD_VARIATION,  # TODO: 'hot'-fix for BigJob
                          api.PROCESSES_PER_HOST,
                          api.TOTAL_CPU_COUNT,
                          api.TOTAL_GPU_COUNT,
                          api.NUMBER_OF_PROCESSES,
                          api.JOB_CONTACT],
    "job_attributes":    [api.EXIT_CODE,
                          api.EXECUTION_HOSTS,
                          api.CREATED,
                          api.STARTED,
                          api.FINISHED],
    "metrics":           [api.STATE],
    "callbacks":         [api.STATE],
    "contexts":          {"ssh": "SSH public/private keypair",
                          "x509": "GSISSH X509 proxy context",
                          "userpass": "username/password pair (ssh)"}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC = {
    "name":          _ADAPTOR_NAME,
    "cfg_options":   _ADAPTOR_OPTIONS,
    "capabilities":  _ADAPTOR_CAPABILITIES,
    "description":  """
The Cobalt adaptor allows to run and manage jobs on
`Cobalt <http://trac.mcs.anl.gov/projects/cobalt>`_controlled HPC clusters.
""",
    "example": "examples/jobs/cobaltjob.py",
    "schemas": {"cobalt":        "connect to a local cluster",
                "cobalt+ssh":    "connect to a remote cluster via SSH",
                "cobalt+gsissh": "connect to a remote cluster via GSISSH"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA
#
_ADAPTOR_INFO = {
    "name"        : _ADAPTOR_NAME,
    "version"     : "v0.1",
    "schemas"     : _ADAPTOR_SCHEMAS,
    "capabilities": _ADAPTOR_CAPABILITIES,
    "cpis"        : [
                        {
                            "type": "radical.saga.job.Service",
                            "class": "CobaltJobService"
                        },
                        {
                            "type": "radical.saga.job.Job",
                            "class": "CobaltJob"
                        }
                    ]
}


###############################################################################

# The adaptor class
#
class Adaptor (a_base.Base):
    """ this is the actual adaptor class, which gets loaded by SAGA (i.e. by
        the SAGA engine), and which registers the CPI implementation classes
        which provide the adaptor's functionality.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        a_base.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        self.id_re = re.compile('^\[(.*)\]-\[(.*?)\]$')
        self.epoch = datetime.datetime(1970,1,1)

        # Adaptor Options
        self.base_workdir     = self._cfg['base_workdir']
        self.purge_on_start   = self._cfg['purge_on_start']
        self.purge_older_than = self._cfg['purge_older_than']

        self.base_workdir     = os.path.normpath(self.base_workdir)

        # dictionaries to keep track of certain Cobalt jobs data
        self._script_file         = dict()  # location of cobalt script file
        self._job_current_workdir = dict()  # working dir, for status checking


    # --------------------------------------------------------------------------
    #
    def sanity_check(self):
        # FIXME: also check for gsissh
        pass

    # --------------------------------------------------------------------------
    #
    def parse_id(self, id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(id)

        if not match or len(match.groups()) != 2:
            raise rse.BadParameter("Cannot parse job id '%s'" % id)

        return (match.group(1), match.group(2))


###############################################################################
# CobaltJobService
class CobaltJobService (cpi_job.Service):
    """ implements cpi_job.Service
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        self._mt  = None
        _cpi_base = super(CobaltJobService, self)
        _cpi_base.__init__(api, adaptor)

        self._adaptor = adaptor

    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self.close()


    # --------------------------------------------------------------------------
    #
    def close(self):

        if  self.mt :
            self.mt.stop()
            self.mt.join(10)  # don't block forever on join()

        self._logger.info("Job monitoring thread stopped.")

        self.finalize(True)


    # --------------------------------------------------------------------------
    #
    def finalize(self, kill_shell=False):

        if  kill_shell :
            if  self.shell :
                self.shell.finalize (True)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        """ service instance constructor
        """
        self.rm      = rm_url
        self.session = session
        self.ppn     = 64       # DEFAULT Theta
        self.is_cray = ""
        self.queue   = None
        self.shell   = None
        self.jobs    = dict()
        self.gres    = None

        # the monitoring thread - one per service instance
        self.mt = _job_state_monitor(job_service=self)
        self.mt.start()

        rm_scheme = rm_url.scheme
        pty_url   = Url(rm_url)

        # this adaptor supports options that can be passed via the
        # 'query' component of the job service URL.
        if rm_url.query:
            for key, val in parse_qs(rm_url.query).items():
                if key == 'queue':
                    self.queue = val[0]
                # Disableing 'ppn' since Blue Gene/Q only supports 16 PPN
                # elif key == 'ppn':
                #     self.ppn = int(val[0])

        # we need to extract the scheme for PTYShell. That's basically the
        # job.Service Url without the pbs+ part. We use the PTYShell to execute
        # pbs commands either locally or via gsissh or ssh.
        if rm_scheme == "cobalt":
            pty_url.scheme = "fork"
        elif rm_scheme == "cobalt+ssh":
            pty_url.scheme = "ssh"
        elif rm_scheme == "cobalt+gsissh":
            pty_url.scheme = "gsissh"

        # these are the commands that we need in order to interact with Cobalt.
        # the adaptor will try to find them during initialize(self) and bail
        # out in case they are note available.
        self._commands = {'nodelist': None,
                          'partlist': None,
                          'qstat':    None,
                          'qsub':     None,
                          'qdel':     None,
                        # 'runjob':   None  # For running scripts
                          }

        self.shell = rsups.PTYShell(pty_url, self.session)

        # self.shell.set_initialize_hook(self.initialize)
        # self.shell.set_finalize_hook(self.finalize)

        self.initialize()
        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # Create the staging directory
        ret, out, _ = self.shell.run_sync ("mkdir -p %s"
                                                   % self._adaptor.base_workdir)
        if  ret != 0 :
            raise rse.NoSuccess("Error creating staging directory. (%s): (%s)"
                           % (ret, out))

        # Purge temporary files
        if self._adaptor.purge_on_start:
            cmd = "find %s -type f -mtime +%d -print -delete | wc -l" % (
                self._adaptor.base_workdir,
                self._adaptor.purge_older_than
            )
            ret, out, _ = self.shell.run_sync(cmd)
            if ret == 0 and out != "0":
                self._logger.info("Purged %s temporary files" % out)

        # Check if all required cobalt tools are available
        for cmd in list(self._commands.keys()):

            ret, out, _ = self.shell.run_sync("which %s " % cmd)

            if ret != 0:
                raise rse.NoSuccess("Error finding Cobalt tools: %s" % out)

            else:
                path = out.strip()  # strip removes newline
                ret, out, _ = self.shell.run_sync("%s --version" % cmd)

                if ret != 0:
                    raise rse.NoSuccess("Error finding Cobalt tools: %s" % out)

                else:
                    # version is reported as
                    # "version: x.y.z" #.strip().split()[1]
                    version = out.strip().replace('\n', '')

                    # add path and version to the command dictionary
                    self._commands[cmd] = {"path": path, "version": version}
        self._logger.info("Found Cobalt tools: %s" % self._commands)


    # --------------------------------------------------------------------------
    #
    def _job_run(self, job_obj):
        """ runs a job via qsub
        """

        # Defaults ...
        cobalt_script_file      = None
        job_current_workdir     = '$HOME'

        # get the job description
        jd = job_obj.get_description()

        # normalize working directory path
        if  jd.working_directory :
            jd.working_directory = os.path.normpath (jd.working_directory)

        # TODO: Why would one want this?
        if self.queue and jd.queue:
            self._logger.warning("Job service was instantiated explicitly with \
                'queue=%s', but job description tries to a different queue: \
                '%s'. Using '%s'." %
                                (self.queue, jd.queue, self.queue))

        try:
            # create a Cobalt job script from SAGA job description
            script = _cobaltscript_generator(url=self.rm,
                                   logger=self._logger,
                                   jd=jd,
                                   ppn=self.ppn,
                                   queue=self.queue,
                                   is_cray=False,
                                   run_job=self._commands['qsub']['path']
                                  )
            self._logger.info("Generated Cobalt script: %s" % str(script))
        except Exception as e:
            raise rse.BadParameter('cobalt script generation failed') from e

        # try to create the working directory (if defined)
        # WARNING: this assumes a shared filesystem between login node and
        #          compute nodes.
        if jd.working_directory:
            self._logger.info("Create workdir %s" % jd.working_directory)
            ret, out, _ = self.shell.run_sync("mkdir -p %s"
                                                       % (jd.working_directory))
            job_current_workdir = jd.working_directory  # Keep track of the cwd,
            if ret != 0:
                # something went wrong
                raise rse.NoSuccess("Couldn't create workdir - %s" % out)


        # Now we want to execute the script. This process consists of two steps:
        # (1) we create a temporary file with 'mktemp'
        #     in the 'self._adaptor.base_workdir'
        #
        # (2) write the contents of
        #     the generated Cobalt script into it, remove the first empty line
        #     and make sure it is executable
        #
        # (3) we call 'qsub --mode script <tmpfile>' to submit the script to
        #     the queueing system
        #
        self._logger.info("Creating Cobalt script file at %s"
                         % jd.working_directory)
        ret, out, _ = self.shell.run_sync("SCRIPTFILE=`mktemp -p %s \
                -t RS-PBSProJobScript.XXXXXX` && echo $SCRIPTFILE"
                % (self._adaptor.base_workdir))
        if ret != 0:
            raise rse.NoSuccess("Couldn't create Cobalt script file - %s" % out)

        # Save Script file for later...
        # Cobalt *needs* the file to stick around, even after submission
        # so, we will keep the file around and delete it *only* when
        # the job is done.
        cobalt_script_file = out.strip()

        self._logger.info("script file: %s" % cobalt_script_file)
        ret, out, _ = self.shell.run_sync('SCRIPTFILE="%s" && echo "%s" \
                > $SCRIPTFILE && echo "$(tail -n +2 $SCRIPTFILE)" \
                > $SCRIPTFILE && chmod +x $SCRIPTFILE && echo $SCRIPTFILE'
                % (cobalt_script_file, script))
        if ret != 0:
            raise rse.NoSuccess("Couldn't create Cobalt script file - %s" % out)

        cmdline = "%s --mode script %s" \
                % (self._commands['qsub']['path'], cobalt_script_file)
        ret, out, _ = self.shell.run_sync(cmdline)

        if ret != 0:
            # something went wrong
            raise rse.NoSuccess(
                "\nError message while running 'qsub' (%s): %s\nCommand: %s\n" %
                (ret, out, cmdline))

        else:
            # parse the job id. qsub usually returns just the job id, but
            # sometimes there are a couple of lines of warnings before.
            # if that's the case, we log those as 'warnings'
            lines = out.split('\n')
            lines = [lines for lines in lines if lines != '']  # remove empty

            if len(lines) > 1:
                self._logger.warning('qsub: %s' % ''.join(lines[:-2]))

            # we asssume job id is in the last line
            job_id = "[%s]-[%s]" % (self.rm, lines[-1].strip().split('.')[0])
            self._logger.info("Submitted Cobalt job with id: %s" % job_id)

            state = api.PENDING

            # populate job info dict
            self.jobs[job_id] = {'obj'         : job_obj,
                                 'job_id'      : job_id,
                                 'state'       : state,
                                 'exec_hosts'  : None,
                                 'returncode'  : None,
                                 'create_time' : None,
                                 'start_time'  : None,
                                 'end_time'    : None,
                                 'gone'        : False
                                 }

            self._logger.info("assign job id  %s / %s / %s to watch list (%s)"
                % (None, job_id, job_obj, list(self.jobs.keys())))

            # set status to 'pending' and manually trigger callback
            job_obj._attributes_i_set('state', state, job_obj._UP, True)

            # Since we now have the job_id, lets track the job's current
            # workdir and script file.  We do this in the adaptor
            self._adaptor._job_current_workdir[job_id]  = job_current_workdir
            self._adaptor._script_file[job_id]          = cobalt_script_file

            # return the job id
            return job_id


    # --------------------------------------------------------------------------
    #
    def _retrieve_job(self, job_id):
        """ see if we can get some info about a job that we don't
            know anything about
        """
        pass


    # --------------------------------------------------------------------------
    #
    def _job_get_info(self, job_id, reconnect):
        """ Get job information attributes via qstat.
        """

        # If we don't have the job in our dictionary, we don't want it,
        # unless we are trying to reconnect.
        if not reconnect and job_id not in self.jobs:
            raise rse.NoSuccess("Unknown job id: %s." % job_id)

        if not reconnect:
            # job_info contains the info collect when _job_get_info
            # was called the last time
            job_info = self.jobs[job_id]

            # if the 'gone' flag is set, there's no need to query the job
            # state again. it's gone forever
            if job_info['gone'] is True:
                return job_info
        else:
            # Create a template data structure
            job_info = {
                'job_id':       job_id,
                'state':        api.UNKNOWN,
                'job_name':     None,
                'exec_hosts':   None,
                'returncode':   None,
                'create_time':  None,
                'start_time':   None,
                'end_time':     None,
                'gone':         False
            }

        rm, pid = self._adaptor.parse_id(job_id)

        # run the Cobalt 'qstat' command to get some info about our job
        qstat_flag = '--full --long'
        ret, out, _ = self.shell.run_sync("unset GREP_OPTIONS; %s %s %s"
                      "| grep -E -i -e '^ *JobName ' "
                                  " -e '^ *QueuedTime ' "
                                  " -e '^ *RunTime ' "
                                  " -e '^ *Nodes ' "
                                  " -e '^ *Procs ' "
                                  " -e '^ *State ' "
                                  " -e '^ *Location ' "
                                  " -e '^ *StartTime ' "
                                  " -e '^ *SubmitTime ' "
                                  " -e '^ *S '"
                % (self._commands['qstat']['path'], qstat_flag, pid))

        if ret != 0:

            if reconnect:
                raise rse.NoSuccess("Couldn't reconnect to job '%s': %s"
                                   % (job_id, out))

            if out.strip() == '':
                # Cobalt's 'qstat' command return's nothing
                # When a job is finished but it exists with code '1'
                # Let's see the job's final state in the job's 'cobaltlog' file
                # which can be found at
                #        'self._adaptor._job_current_workdir_cwd/pid.cobaltlog'
                # If file is found: get the final status of the job
                # If file not found: let's assume it FAILED.

                # Run a 'cat' command to the final info about our job
                # Sample OUTPUT:
                # ...
                # Mon Jan 23 02:44:05 2017 +0000 (UTC) \
                # Info: task completed normally with an exit code of 126;\
                #       initiating job cleanup and removal

                ret, out, _ = self.shell.run_sync("unset GREP_OPTIONS; "
                              "cat %s/%s.cobaltlog "
                              "| grep -P -i '^[A-Z][a-z]{2} [A-Z][a-z]{2}"
                                             " \d{2} \d{2}:\d{2}:\d{2} \d{4}"
                                             " \+\d{4} \([A-Za-z]+\) *Info:"
                                             " task completed'"
                    % (self._adaptor._job_current_workdir[job_id], pid))

                if ret != 0:
                    if reconnect:
                        raise rse.NoSuccess("Couldn't reconnect to job '%s': %s"
                                            % (job_id, out))
                elif out.strip() == '':
                    # Let's see if the last known job state was running or pending. in
                    # that case, the job is gone now, which can either mean DONE,
                    # or FAILED. the only thing we can do is set it to 'DONE'
                    job_info['gone'] = True
                    # TODO: we can also set the end time?
                    self._logger.warning("Previously running job has disappeared. "
                            "This probably means that the backend doesn't store "
                            "information about finished jobs. Setting state to 'DONE'.")
                    if job_info['state'] in [api.RUNNING, api.PENDING]:
                        job_info['state'] = api.DONE
                    else:
                        # TODO: This is an uneducated guess?
                        job_info['state'] = api.FAILED
                else:
                    try:
                        # Found the cobaltlot file, let's grab the result...
                        matches = re.search('^([A-Z][a-z]{2} [A-Z][a-z]{2} \d{2} \d{2}:\d{2}:\d{2} \d{4} \+\d{4} \([A-Za-z]+\)) *Info: task completed .+ an exit code of (\d+);', out)
                        timestamp = matches.group(1).strip()
                        exit_code = matches.group(2).strip()
                    except Exception as e:
                        raise rse.NoSuccess('Could not parse job status') from e

                    # Current format: Mon Jan 23 02:44:05 2017 +0000 (UTC)
                    # ASSUMPTION: Date is in UTC (as seen on the servers)
                    # Will be parsed as UTC and output format:
                    # DDD mmm dd HH:MM:SS YYYY +0000 (UTC)
                    # Wed Dec 21 15:51:34 2016 +0000 (UTC)
                    end_time = datetime.datetime.strptime(timestamp,
                                             "%a %b %d %H:%M:%S %Y +0000 (UTC)")
                    job_info['end_time'] = (end_time - self._adaptor.epoch) \
                                                                .total_seconds()

                    # Return code is on position '13'
                    job_info['returncode'] = int(exit_code)

                    # Final Job State given the exit code
                    if job_info['returncode'] != 0:
                        job_info['state'] = api.FAILED
                    else:
                        job_info['state'] = api.DONE
            else:
                # something went wrong
                raise rse.NoSuccess("Error retrieving job info via 'qstat': %s"
                                    % out)
        else:

            # The job seems to exist on the system. let's process some data.

            # TODO: make the parsing "contextual", in the sense that it takes
            #       the state into account.

            # parse the egrep result. this should look something like this:
            #       QueuedTime        : 00:00:04
            #       RunTime           : 00:00:39
            #       Nodes             : 2
            #       State             : running
            #       Procs             : 3
            #       Location          : CENTOS-04000-37331-512
            #       StartTime         : Tue Nov 29 02:11:45 2016 +0000 (UTC)
            #       SubmitTime        : Tue Nov 29 02:11:40 2016 +0000 (UTC)
            #       S                 : R
            results = out.split('\n')
            for line in results:
                if len(line.split(':')) == 2:
                    key, val = line.split(':')
                    key = key.strip()
                    val = val.strip()

                    # The ubiquitous job state
                    if key in ['S']:  # Cobalt's PBS-like state
                        job_info['state'] = _cobalt_to_saga_jobstate(val)

                    # Hosts where the job ran
                    elif key in ['Location']:  # Cobalt's Node/Partition
                        # format CENTOS-04000-37331-512
                        job_info['exec_hosts'] = val

                    # Time job got created in the queue
                    elif key in ['SubmitTime']:
                        job_info['create_time'] = val

                    # Time job started to run
                    elif key in ['StartTime']:
                        job_info['start_time'] = val

                    # Job name
                    elif key in ['JobName']:
                        job_info['job_name'] = val


        # return the updated job info
        return job_info


    # --------------------------------------------------------------------------
    #
    def _parse_qstat(self, haystack, job_info):
        # return the new job info dict
        return job_info


    # --------------------------------------------------------------------------
    #
    def _job_get_state(self, job_id):
        """ get the job's state
        """
        return self.jobs[job_id]['state']


    # --------------------------------------------------------------------------
    #
    def _job_get_exit_code(self, job_id):
        """ get the job's exit code
        """
        ret = self.jobs[job_id]['returncode']

        # FIXME: 'None' should cause an exception
        if ret is None : return None
        else           : return int(ret)


    # --------------------------------------------------------------------------
    #
    def _job_get_execution_hosts(self, job_id):
        """ get the job's exit code
        """
        return self.jobs[job_id]['exec_hosts']


    # --------------------------------------------------------------------------
    #
    def _job_get_create_time(self, job_id):
        """ get the job's creation time
        """
        return self.jobs[job_id]['create_time']


    # --------------------------------------------------------------------------
    #
    def _job_get_start_time(self, job_id):
        """ get the job's start time
        """
        return self.jobs[job_id]['start_time']


    # --------------------------------------------------------------------------
    #
    def _job_get_end_time(self, job_id):
        """ get the job's end time
        """
        return self.jobs[job_id]['end_time']


    # --------------------------------------------------------------------------
    #
    def _job_cancel(self, job_id):
        """ cancel the job via 'qdel'
        """
        rm, pid = self._adaptor.parse_id(job_id)

        ret, out, _ = self.shell.run_sync("%s %s\n"
                                        % (self._commands['qdel']['path'], pid))

        if ret != 0:
            raise rse.NoSuccess("Error canceling job via 'qdel': %s" % out)

        # assume the job was succesfully canceled
        self.jobs[job_id]['state'] = api.CANCELED


    # --------------------------------------------------------------------------
    #
    def _job_wait(self, job_id, timeout):
        """ wait for the job to finish or fail
        """
        time_start = time.time()
        time_now   = time_start
        rm, pid    = self._adaptor.parse_id(job_id)

        while True:
            state = self.jobs[job_id]['state']  # this gets updated in the bg.

            if state in [api.DONE, api.FAILED, api.CANCELED]:
                return True

            # avoid busy poll
            time.sleep(SYNC_WAIT_UPDATE_INTERVAL)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        """ implements cpi_job.Service.get_url()
        """
        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         "job_description": jd,
                         "job_schema":      self.rm.schema,
                         "reconnect":       False
                         }

        # create and return a new job object
        return api.Job(_adaptor=self._adaptor,
                       _adaptor_state=adaptor_state)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, job_id):
        """ Implements cpi_job.Service.get_job()

            Re-create job instance from a job-id.
        """

        # If we already have the job info, we just pass the current info.
        if job_id in self.jobs :
            return self.jobs[job_id]['obj']

        # Try to get some initial information about this job (again)
        job_info = self._job_get_info(job_id, reconnect=True)

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         # TODO: fill job description
                         "job_description": api.Description(),
                         "job_schema":      self.rm.schema,
                         "reconnect":       True,
                         "reconnect_jobid": job_id
                         }

        job_obj = api.Job(_adaptor=self._adaptor,
                          _adaptor_state=adaptor_state)

        # throw it into our job dictionary.
        job_info['obj']   = job_obj
        self.jobs[job_id] = job_info

        return job_obj


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        """ implements cpi_job.Service.get_url()
        """
        return self.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        """ implements cpi_job.Service.list()
        """
        ids = []

        ret, out, _ = self.shell.run_sync("unset GREP_OPTIONS; %s "
            "--header=JobID:JobName:User:Walltime:RunTime:Nodes:State:"
            "short_state:Location:Queue | grep `whoami`"
            % self._commands['qstat']['path'])

        if ret != 0 and len(out) > 0:
            raise rse.NoSuccess("failed to list jobs via 'qstat': %s" % out)

        elif ret != 0 and len(out) == 0:
            # qstat | grep `` exits with 1 if the list is empty
            pass

        else:
            for line in out.split("\n"):
                # output looks like this:
                # (not inlcuding the first line - that is for reference only)
                #
                # JobID  JobName  User     Walltime  RunTime   Nodes  State      S  Location                Queue
                # 32     hello    vagrant  00:50:00  00:00:00  2      starting   R  CENTOS-04400-37731-512  default
                # 33     hello    vagrant  00:50:00  00:00:11  2      running    R  CENTOS-04440-37771-512  default
                # 35     hello    vagrant  00:50:00  N/A       4      queued     Q  None                    default
                # 36     hello    vagrant  00:50:00  N/A       2      queued     Q  None                    default
                # 37     hello    vagrant  00:50:00  00:00:22  2      running    R  CENTOS-04040-37371-512  default
                # 38     hello    vagrant  00:50:00  N/A       2      queued     Q  None                    default
                # 40     hello    vagrant  00:50:00  N/A       2      user_hold  H  None                    default
                # 41     hello    vagrant  00:50:00  N/A       1      queued     Q  None                    default
                if len(line.split()) > 1:
                    job_id = "[%s]-[%s]" % (self.rm, line.split()[0].strip())
                    ids.append(str(job_id))
        return ids


    # --------------------------------------------------------------------------
    #
    def container_run (self, jobs) :
        self._logger.debug ("container run: %s"  %  str(jobs))
        # TODO: this is not optimized yet
        for job in jobs:
            job.run ()


    # --------------------------------------------------------------------------
    #
    def container_wait (self, jobs, mode, timeout) :
        self._logger.debug ("container wait: %s"  %  str(jobs))
        # TODO: this is not optimized yet
        for job in jobs:
            job.wait ()


    # --------------------------------------------------------------------------
    #
    def container_cancel (self, jobs, timeout) :
        # TODO: this is not optimized yet
        for job in jobs:
            job.cancel(timeout)


    # --------------------------------------------------------------------------
    #
    def container_get_states(self, jobs):
        # TODO: this is not optimized yet
        states = list()
        for job in jobs:
            states.append(job.get_state())
        return states


###############################################################################
#
class CobaltJob (cpi_job.Job):
    """ implements cpi_job.Job
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        # initialize parent class
        _cpi_base = super(CobaltJob, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    def _get_impl(self):
        return self


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, job_info):
        """ implements cpi_job.Job.init_instance()
        """
        # init_instance is called for every new rs.job.Job object
        # that is created
        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        if job_info['reconnect'] is True:
            self._id      = job_info['reconnect_jobid']
            self._name    = self.jd.get(api.NAME)
            self._started = True
        else:
            self._id      = None
            self._name    = self.jd.get(api.NAME)
            self._started = False

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        """ implements cpi_job.Job.get_state()
        """
        if  self._started is False:
            return api.NEW

        return self.js._job_get_state(job_id=self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):
        """ implements cpi_job.Job.wait()
        """
        if self._started is False:
            raise rse.IncorrectState("Can't wait for job that hasn't been started")
        else:
            self.js._job_wait(job_id=self._id, timeout=timeout)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):
        """ implements cpi_job.Job.cancel()
        """
        if self._started is False:
            raise rse.IncorrectState("Can't wait for job that hasn't been started")

        else:
            self.js._job_cancel(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):
        """ implements cpi_job.Job.run()
        """
        self._id = self.js._job_run(self._api())
        self._started = True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):
        """ implements cpi_job.Job.get_service_url()
        """
        return self.js.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):
        """ implements cpi_job.Job.get_id()
        """
        return self._id


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name (self):
        """ Implements cpi_job.Job.get_name() """
        return self._name


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):
        """ implements cpi_job.Job.get_exit_code()
        """
        if not self._started:
            return None
        else:
            return self.js._job_get_exit_code(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):
        """ implements cpi_job.Job.get_created()
        """
        if self._started is False:
            return None
        else:
            # FIXME: convert to EPOCH
            return self.js._job_get_create_time(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):
        """ implements cpi_job.Job.get_started()
        """
        if self._started is False:
            return None
        else:
            # FIXME: convert to EPOCH
            return self.js._job_get_start_time(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):
        """ implements cpi_job.Job.get_finished()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_end_time(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):
        """ implements cpi_job.Job.get_execution_hosts()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_execution_hosts(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description(self):
        """ implements cpi_job.Job.get_execution_hosts()
        """
        return self.jd


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.saga.adaptors.cpi.decorators as cpi_dec
import radical.saga.adaptors.cpi.namespace  as cpi_ns
import radical.saga.adaptors.cpi.attributes as cpi_att

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


# keep order of inheritance!  super() below uses MRO
class LogicalDirectory (cpi_ns.directory.Directory, 
                        cpi_att.Attributes) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_nsdirec = super  (LogicalDirectory, self)
        self._cpi_nsdirec.__init__ (api, adaptor)

    @SYNC
    def init_instance           (self, url, flags, session)          : pass
    @ASYNC
    def init_instance_async     (self, url, flags, session)          : pass


    @SYNC
    def open                    (self, tgt, flags, ttype)            : pass
    @ASYNC
    def open_async              (self, tgt, flags, ttype)            : pass

    @SYNC
    def open_dir                (self, tgt, flags, ttype)            : pass
    @ASYNC
    def open_dir_async          (self, tgt, flags, ttype)            : pass

    # ----------------------------------------------------------------
    #
    # replica methods
    #
    @SYNC
    def get_size                (self, tgt, ttype)                   : pass
    @ASYNC
    def get_size_async          (self, tgt, ttype)                   : pass

    @SYNC
    def is_file                 (self, tgt, ttype)                   : pass
    @ASYNC
    def is_file_async           (self, tgt, ttype)                   : pass

    @SYNC
    def is_file_self            (self, ttype)                        : pass
    @ASYNC
    def is_file_self_async      (self, ttype)                        : pass

    @SYNC
    def find_replicas       (self, name_pattern, attr_pattern, flags, ttype)  : pass
    @ASYNC
    def find_replicas_async (self, name_pattern, attr_pattern, flags, ttype)  : pass








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Provides the SAGA Resource CPI """

from .. import decorators as cpi_dec
from .. import base       as cpi_base
from .. import sasync     as cpi_async


SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


class Resource (cpi_base.CPIBase, cpi_async.Async) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (Resource, self)
        self._cpi_base.__init__ (api, adaptor)

    @SYNC
    def init_instance             (self, info,    ttype) : pass
    @ASYNC
    def init_instance_async       (self, info,    ttype) : pass


    # ----------------------------------------------------------------
    #
    # resource methods
    #
    @SYNC
    def reconfig           (self, descr,          ttype) : pass
    @ASYNC
    def reconfig_async     (self, descr,          ttype) : pass

    @SYNC
    def destroy            (self,                 ttype) : pass
    @ASYNC
    def destroy_async      (self,                 ttype) : pass

    @SYNC
    def wait               (self, state, timeout, ttype) : pass
    @ASYNC
    def wait_async         (self, state, timeout, ttype) : pass


    #-----------------------------------------------------------------
    # attribute getters
    @SYNC
    def get_id                 (self, ttype)             : pass
    @ASYNC
    def get_id_async           (self, ttype)             : pass

    @SYNC
    def get_rtype              (self, ttype)             : pass
    @ASYNC
    def get_rtype_async        (self, ttype)             : pass

    @SYNC
    def get_state              (self, ttype)             : pass
    @ASYNC
    def get_state_async        (self, ttype)             : pass

    @SYNC
    def get_state_detail       (self, ttype)             : pass
    @ASYNC
    def get_state_detail_async (self, ttype)             : pass

    @SYNC
    def get_access             (self, ttype)             : pass
    @ASYNC
    def get_access_async       (self, ttype)             : pass

    @SYNC
    def get_manager            (self, ttype)             : pass
    @ASYNC
    def get_manager_async      (self, ttype)             : pass

    @SYNC
    def get_description        (self, ttype)             : pass
    @ASYNC
    def get_description_async  (self, ttype)             : pass


class Compute (Resource) : pass
class Storage (Resource) : pass
class Network (Resource) : pass






#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import sys

import radical.saga as rs


def main():

    try:

        c = rs.Context ('ssh')
        c.user_id   = 'tg12736'
        c.user_cert = '/home/user/ssh/id_rsa_xsede' # private key derived from cert

        s = rs.Session (default=False)            # create session with no contexts
        s.add_context (c)

        js = rs.job.Service ('ssh://login1.stampede.tacc.utexas.edu', session=s)
        js.run_job ("/bin/true")

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())




__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


''' Provides a parser class for the file transfer specification as it is
    defined in GFD.90, sction 4.1.3.
'''

from ... import exceptions as rse

# 4.1.3 File Transfer Specifications (GFD90 p 176-177)
#
# The syntax of a file transfer directive for the job description is modeled on
# the LSF syntax (LSF stands for Load Sharing Facility, a commercial job
# scheduler by Platform Computing), and has the general syntax:
# local_file operator remote_file
# Both the local_file and the remote_file can be URLs. If they are not URLs,
# but full or relative pathnames, then the local_file is relative to the host
# where the submission is executed, and the remote_file is evaluated on the
# execution host of the job. The operator is one of the following four:
#
# '>  copies the local file to the remote file before the job starts.
#      Overwrites the remote file if it exists.
# '>>' copies the local file to the remote file before the job starts.
#      Appends to the remote file if it exists.
# '<'  copies the remote file to the local file after the job finishes.
#      Overwrites the local file if it exists.
# '<<' copies the remote file to the local file after the job finishes.
#      Appends to the local file if it exists.


class TransferDirectives(object):

    def __init__(self, directives_list):

        self._in_overwrite = dict()
        self._in_append = dict()
        self._out_overwrite = dict()
        self._out_append = dict()

        # each line in directives_list should contain one directive
        for directive in directives_list:
            if (directive.count('>') > 2) or (directive.count('<') > 2):
                msg = "'%s' is not a valid transfer directive string."
                raise rse.BadParameter(msg)
            elif '<<' in directive:
                (local, remote) = directive.split('<<')
                self._out_append[local.strip()] = remote.strip()
            elif '>>' in directive:
                (local, remote) = directive.split('>>')
                self._in_append[local.strip()] = remote.strip()
            elif '<' in directive:
                (local, remote) = directive.split('<')
                self._out_overwrite[local.strip()] = remote.strip()
            elif '>' in directive:
                (local, remote) = directive.split('>')
                self._in_overwrite[local.strip()] = remote.strip()
            else:
                msg = "'%s' is not a valid transfer directive string." % directive
                raise rse.BadParameter(msg)

    def _dicts_to_string_list(self):
        slist = list()
        for (local, remote) in self._in_overwrite.items():
            slist.append('%s > %s' % (local, remote))
        for (local, remote) in self._in_append.items():
            slist.append('%s >> %s' % (local, remote))
        for (remote, local) in self._out_overwrite.items():
            slist.append('%s < %s' % (local, remote))
        for (remote, local) in self._out_append.items():
            slist.append('%s << %s' % (local, remote))
        return slist

    def __str__(self):
        """ String representation.
        """
        return str(self._dicts_to_string_list())

    @property
    def in_overwrite_dict(self):
        return self._in_overwrite

    @property
    def in_append_dict(self):
        return self._in_append

    @property
    def out_overwrite_dict(self):
        return self._out_overwrite

    @property
    def out_append_dict(self):
        return self._out_append

    @property
    def string_list(self):
        return self._dicts_to_string_list()





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils.signatures as rus

from .. import attributes  as sa
from .  import constants   as c


# ------------------------------------------------------------------------------
#
class Message (sa.Attributes) :


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Message', 
                  rus.optional (int, rus.nothing))  # size
    @rus.returns (rus.nothing)
    def __init__ (self, size=None) : 
        '''
        size:      expected size of buffer (informative, not normative)
        ret:       Message
        '''

        # set attribute interface properties
        self._attributes_camelcasing   (True)
        self._attributes_allow_private (True)
        self._attributes_extensible    (True, getter=self._attribute_getter, 
                                              setter=self._attribute_setter,
                                              lister=self._attribute_lister,
                                              caller=self._attribute_caller)

        # register properties with the attribute interface 
        self._attributes_register(c.ID,     None, sa.STRING, sa.SCALAR,
                                                             sa.READONLY)
        self._attributes_register(c.SENDER, None, sa.STRING, sa.SCALAR,
                                                             sa.READONLY)


    # --------------------------------------------------------------------------
    #
    # class methods
    #
    @rus.takes   ('Message')
    @rus.returns (rus.anything)
    def get_id (self) :

        return self._adaptor.get_id ()


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Message')
    @rus.returns (rus.anything)
    def get_sender (self) :

        return self._adaptor.get_sender ()


    # --------------------------------------------------------------------------
    #
    # attribute methods
    #
    @rus.takes   ('Message', 
                  str)
    @rus.returns (rus.anything)
    def _attribute_getter (self, key) :

        return self._adaptor.attribute_getter (key)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Message', 
                  str,
                  rus.anything)
    @rus.returns (rus.nothing)
    def _attribute_setter (self, key, val) :

        return self._adaptor.attribute_setter (key, val)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Message')
    @rus.returns (rus.list_of (rus.anything))
    def _attribute_lister (self) :

        return self._adaptor.attribute_lister ()


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Message', 
                  str, 
                  int, 
                  callable)
    @rus.returns (rus.anything)
    def _attribute_caller (self, key, id, cb) :

        return self._adaptor.attribute_caller (key, id, cb)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils            as ru
import radical.utils.signatures as rus

from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base    as sab
from ..namespace import entry   as nsentry

from .. import attributes       as sa
from .. import session          as ss
from .. import task             as st
from .  import constants        as c


# ------------------------------------------------------------------------------
#
class LogicalFile (nsentry.Entry, sa.Attributes) :

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalFile',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, flags=c.READ, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        '''
        __init__(url=None, flags=READ, session=None)

        url:       saga.Url
        flags:     flags enum
        session:   saga.Session
        ret:       obj
        '''

        # param checks
        if not flags : flags = 0
        url = ru.Url (url)

        self._nsentry = super  (LogicalFile, self)
        self._nsentry.__init__ (url, flags, session,
                                _adaptor, _adaptor_state, _ttype=_ttype)


    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('LogicalFile',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create (cls, url=None, flags=c.READ, session=None, ttype=None) :
        '''
        url:       saga.Url
        flags:     saga.replica.flags enum
        session:   saga.Session
        ttype:     saga.task.type enum
        ret:       saga.Task
        '''

        if not flags : flags = 0
        _nsentry = super (LogicalFile, cls)
        return _nsentry.create (url, flags, session, ttype=ttype)


    # ----------------------------------------------------------------
    #
    @rus.takes   ('LogicalFile',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_file (self, ttype=None) :
        '''
        is_file()

        ttype:          saga.task.type enum
        ret:            bool / saga.Task
        '''
        return self.is_entry (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    # replica methods
    #
    @rus.takes   ('LogicalFile',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def get_size (self, ttype=None) :
        '''
        get_size()

        Return the size of the file.

        ttype:    saga.task.type enum
        ret:      int / saga.Task

        Returns the size of the physical file represented by this logical file
        (in bytes)

           Example::

               # get a file handle
               lf = saga.replica.LogicalFile("irods://localhost/tmp/data.bin")

               # print the logical file's size
               print(lf.get_size ())

        '''
        return self._adaptor.get_size_self (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalFile',
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def add_location (self, name, ttype=None) :
        '''
        add_location(name)

        Add a physical location.

        name:           saga.Url
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        '''
        return self._adaptor.add_location (name, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalFile',
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def remove_location (self, name, ttype=None) :
        '''
        remove_location(name)

        Remove a physical location.

        name:           saga.Url
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        '''
        return self._adaptor.remove_location (name, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalFile',
                  rus.optional ((ru.Url, str)),
                  rus.optional ((ru.Url, str)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def update_location (self, old, new, ttype=None) :
        '''
        update_location(old, new)

        Updates a physical location.

        old:            saga.Url
        new:            saga.Url
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        '''
        return self._adaptor.update_location (old, new, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalFile',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (ru.Url), st.Task))
    def list_locations (self, ttype=None) :
        '''
        list_locations()

        List all physical locations of a logical file.

        ttype:          saga.task.type enum
        ret:            list [saga.Url] / saga.Task
        '''
        return self._adaptor.list_locations (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalFile',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def replicate (self, name, flags=None, ttype=None) :
        '''
        replicate(name)

        Replicate a logical file.

        name:           saga.Url
        flags:          flags enum
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        '''
        if not flags : flags = 0
        return self._adaptor.replicate (name, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    # non-GFD.90
    #
    @rus.takes   ('LogicalFile',
                  (ru.Url, str),
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def upload (self, name, tgt=None, flags=None, ttype=None) :
        '''
        upload(name, tgt=None, flags=None)

        Upload a physical file.

        name:           saga.Url
        tgt:            saga.Url
        flags:          flags enum
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        '''
        if not flags : flags = 0
        return self._adaptor.upload (name, tgt, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    # non-GFD.90
    #
    @rus.takes   ('LogicalFile',
                  (ru.Url, str),
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def download (self, name, src=None, flags=None, ttype=None) :
        '''
        download(name, src=None, flags=None)

        Download a physical file.

        name:           saga.Url
        src:            saga.Url
        flags:          flags enum
        ttype:          saga.task.type enum
        ret:            None / saga.Task
        '''
        if not flags : flags = 0
        return self._adaptor.download (name, src, flags, ttype=ttype)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils.signatures as rus

from .adaptors import base   as sab
from . import attributes as sa
from . import base       as sb

from .constants import TYPE,       SERVER,    USER_CERT,   CERT_REPOSITORY
from .constants import USER_PROXY, USER_KEY,  USER_ID,     USER_PASS,   USER_VO
from .constants import LIFE_TIME,  REMOTE_ID, REMOTE_HOST, REMOTE_PORT, TOKEN

# ------------------------------------------------------------------------------
#
class Context (sb.Base, sa.Attributes) :
    '''A SAGA Context object as defined in GFD.90.

    A security context is a description of a security token.  It is important to
    understand that, in general, a context really just *describes* a token, but
    that a context *is not* a token (*). For example, a context may point to
    a X509 certificate -- but it will in general not hold the certificate
    contents.

    Context classes are used to inform the backends used by SAGA on what
    security tokens are expected to be used.  By default, SAGA will be able to
    pick up such tokens from their default location, but in some cases it might
    be necessary to explicitly point to them - then use Session with
    context instances to do so.

    The usage example for contexts is below::

        # define an ssh context
        ctx = saga.Context("SSH")
        ctx.user_cert = '$HOME/.ssh/special_id_rsa'
        ctx.user_key  = '$HOME/.ssh/special_id_rsa.pub'

        # add the context to a session
        session = saga.Session()
        session.add_context(ctx)

        # create a job service in this session -- that job service can now
        # *only* use that ssh context.
        j = saga.job.Service('ssh://remote.host.net/', session=session)


    The Session argument to the job.Service constructor is fully optional
    -- if left out, SAGA will use default session, which picks up some default
    contexts as described above -- that will suffice for the majority of use
    cases.

    ----

    (*) The only exception to this rule is the 'UserPass' key, which is used to
    hold plain-text passwords.  Use this key with care -- it is not good
    practice to hard-code passwords in the code base, or in config files.
    Also, be aware that the password may show up in log files, when debugging or
    analyzing your application.

    '''

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Context',
                  str,
                  rus.optional (sab.Base),
                  rus.optional (dict))
    @rus.returns (rus.nothing)
    def __init__ (self, ctype, _adaptor=None, _adaptor_state={}) :
        '''
        ctype: string
        ret:   None
        '''

        sb.Base.__init__ (self, ctype.lower(), _adaptor, _adaptor_state, ctype, ttype=None)


        from . import attributes as sa

        # set attribute interface propertiesP
        self._attributes_extensible  (False)
        self._attributes_camelcasing (True)

        # register properties with the attribute interface
        self._attributes_register  (TYPE,            None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (SERVER,          None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (TOKEN,           None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (CERT_REPOSITORY, None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (USER_PROXY,      None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (USER_CERT,       None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (USER_KEY,        None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (USER_ID,         None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (USER_PASS,       None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (USER_VO,         None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (LIFE_TIME,       -1,   sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (REMOTE_ID,       None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (REMOTE_HOST,     None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (REMOTE_PORT,     None, sa.STRING, sa.VECTOR, sa.WRITEABLE)

        self.type = ctype


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Context')
    @rus.returns (str)
    def __str__  (self) :

        d = self.as_dict ()
        s = "{"

        for key in sorted (d.keys ()) :
            if  key == 'UserPass' and d[key] :
                s += "'UserPass' : '%s'" % ('x'*len(d[key]))
            else :
                s += "'%s' : '%s'" % (key, d[key])
            s += ', '

        return "%s}" % s[0:-2]


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Context')
    @rus.returns (str)
    def __repr__ (self) :

        return str(self)


    # --------------------------------------------------------------------------
    #
    @rus.takes      ('Context',
                     ('Session', '_DefaultSession'))
    @rus.returns    (rus.nothing)
    def _initialize (self, session) :
        '''
        ret:  None
        '''
        self._adaptor._initialize (session)


    # --------------------------------------------------------------------------
    #
    def __deepcopy__(self, memo):

        ret = Context(self.type)

        for a in self.list_attributes():
            ret.set_attribute(a, self.get_attribute(a))

        return ret


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .. import decorators as cpi_dec
from .. import base       as cpi_base
from .. import sasync     as cpi_async

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


# ------------------------------------------------------------------------------
# keep order of inheritance!  super() below uses MRO
class Entry (cpi_base.CPIBase, cpi_async.Async) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        _cpi_base = super  (Entry, self)
        _cpi_base.__init__ (api, adaptor)

    @SYNC
    def init_instance        (self, url, flags, session)      : pass
    @ASYNC
    def init_instance_async  (self, url, flags, session)      : pass


    # ----------------------------------------------------------------
    #
    # namespace entry methods
    #
    @SYNC
    def close                (self, timeout, ttype)           : pass
    @ASYNC
    def close_async          (self, timeout, ttype)           : pass

    @SYNC
    def get_url              (self, ttype)                    : pass
    @ASYNC
    def get_url_async        (self, ttype)                    : pass

    @SYNC
    def get_cwd              (self, ttype)                    : pass
    @ASYNC
    def get_cwd_async        (self, ttype)                    : pass

    @SYNC
    def get_name             (self, ttype)                    : pass
    @ASYNC
    def get_name_async       (self, ttype)                    : pass

    @SYNC
    def is_dir_self          (self, ttype)                    : pass
    @ASYNC
    def is_dir_self_async    (self, ttype)                    : pass

    @SYNC
    def is_entry_self        (self, ttype)                    : pass
    @ASYNC
    def is_entry_self_async  (self, ttype)                    : pass

    @SYNC
    def is_link_self         (self, ttype)                    : pass
    @ASYNC
    def is_link_self_async   (self, ttype)                    : pass

    @SYNC
    def read_link_self       (self, ttype)                    : pass
    @ASYNC
    def read_link_self_async (self, ttype)                    : pass

    @SYNC
    def copy_self            (self, tgt, flags, ttype)        : pass
    @ASYNC
    def copy_self_async      (self, tgt, flags, ttype)        : pass

    @SYNC
    def link_self            (self, tgt, flags, ttype)        : pass
    @ASYNC
    def link_self_async      (self, tgt, flags, ttype)        : pass

    @SYNC
    def move_self            (self, tgt, flags, ttype)        : pass
    @ASYNC
    def move_self_async      (self, tgt, flags, ttype)        : pass

    @SYNC
    def remove_self          (self, flags, ttype)             : pass
    @ASYNC
    def remove_self_async    (self, flags, ttype)             : pass


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from ..namespace import constants as ns

# filesystem flags enum:
OVERWRITE      = ns.OVERWRITE       #      1
RECURSIVE      = ns.RECURSIVE       #      2
DEREFERENCE    = ns.DEREFERENCE     #      4
CREATE         = ns.CREATE          #      8
EXCLUSIVE      = ns.EXCLUSIVE       #     16
LOCK           = ns.LOCK            #     32
CREATE_PARENTS = ns.CREATE_PARENTS  #     64
TRUNCATE       =                         128
# APPEND       = reserved           #    256
READ           =                         512
WRITE          =                        1024
READ_WRITE     =                        1536
# BINARY       = reserved           #   2048


# attributes
ATTRIBUTE      = 'Attribute'
OBJECT         = 'Object'
TTL            = 'TTL'
CHANGE         = 'Change'
NEW            = 'New'
DELETE         = 'Delete'








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Provides the SAGA Resource CPI """

from .. import decorators as cpi_dec
from .. import base       as cpi_base
from .. import sasync     as cpi_async

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


class Manager (cpi_base.CPIBase, cpi_async.Async) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (Manager, self)
        self._cpi_base.__init__ (api, adaptor)

    @SYNC
    def init_instance              (self, url, session)        : pass
    @ASYNC
    def init_instance_async        (self, url, session)        : pass



    # ----------------------------------------------------------------
    #
    # resource manager methods
    #
    @SYNC
    def list                       (self, type,      ttype)    : pass
    @ASYNC
    def list_async                 (self, type,      ttype)    : pass

    @SYNC
    def get_description            (self, id,        ttype)    : pass
    @ASYNC
    def get_description_async      (self, id,        ttype)    : pass

    @SYNC
    def list_templates             (self, type,      ttype)    : pass
    @ASYNC
    def list_templates_async       (self, type,      ttype)    : pass

    @SYNC
    def get_template               (self, name,      ttype)    : pass
    @ASYNC
    def get_template_async         (self, name,      ttype)    : pass

    @SYNC
    def list_images                (self, type,      ttype)    : pass
    @ASYNC
    def list_images_async          (self, type,      ttype)    : pass

    @SYNC
    def get_image                  (self, name,      ttype)    : pass
    @ASYNC
    def get_image_async            (self, name,      ttype)    : pass

    @SYNC
    def acquire                    (self, descr,     ttype)    : pass
    @ASYNC
    def acquire_async              (self, descr,     ttype)    : pass

    @SYNC
    def destroy                    (self, id,        ttype)    : pass
    @ASYNC
    def destroy_async              (self, id,        ttype)    : pass





